# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.

"""gtfsdbloader - GTFS to GTFS' conversion tool and database loader

Usage:
  gtfsdbloader <database> (--load=<gtfs> | --delete | --list) [--id=<id>]
                        [--logsql] [--lenient] [--schema=<schema>]
                        [--disablenormalize]
  gtfsdbloader (-h | --help)
  gtfsdbloader --version

Options:
  <database>           The database to use. If a file, assume SQLite.
                       For PostgreSQL: "postgresql://user:pwd@host:port/db".
  --load=<gtfs>        The zipped GTFS file to load.
  --delete             Delete feed.
  --list               List all feeds.
  --id=<id>            Set the feed ID in case multiple GTFS are to be loaded.
  -h --help            Show help on options.
  --version            Show lib / program version.
  --logsql             Enable SQL logging (very verbose)
  --lenient            Allow some level of brokenness in GTFS input.
  --schema=<schema>    Set the schema to use (for PostgreSQL).
  --disablenormalize   Disable shape and stop times normalization. Be careful
                       if you use this option, as missing stop times will not
                       be interpolated, and shape_dist_traveled will not be
                       computed or converted to meters.

Examples:
  gtfsdbloader db.sqlite --load=sncf.zip --id=sncf
        Load the GTFS sncf.zip into db.sqlite using id "sncf",
        deleting previous data.
  gtfsdbloader db.sqlite --delete --id=moontransit
        Delete the "moontransit" feed from the database.
  gtfsdbloader db.sqlite --list
        List all feed IDs from db.sqlite
  gtfsdbloader postgresql://gtfs@localhost/gtfs --load gtfs.zip
        Load gtfs.zip into a postgresql database,
        using a default (empty) feed ID.

Authors:
  CEREMA / AFIMB
  Laurent GRÉGOIRE (MECATRAN) <laurent.gregoire@mecatran.com>
"""
from docopt import docopt
from logging import StreamHandler
import logging
import sys
from gtfslib.dao import Dao
import gtfslib

def main():
    arguments = docopt(__doc__, version='gtfsdbloader %s' % gtfslib.__version__)
    if arguments['--id'] is None:
        arguments['--id'] = ""

    # TODO Configure logging properly?
    logger = logging.getLogger('libgtfs')
    logger.setLevel(logging.INFO)
    logger.addHandler(StreamHandler(sys.stdout))

    dao = Dao(arguments['<database>'],
              sql_logging=arguments['--logsql'],
              schema=arguments['--schema'])

    if arguments['--list']:
        for feed in dao.feeds():
            print(feed.feed_id if feed.feed_id != "" else "(default)")

    if arguments['--delete'] or arguments['--load']:
        feed_id = arguments['--id']
        existing_feed = dao.feed(feed_id)
        if existing_feed:
            logger.warn("Deleting existing feed ID '%s'" % feed_id)
            dao.delete_feed(feed_id)
            dao.commit()

    if arguments['--load']:
        dao.load_gtfs(arguments['--load'],
                      feed_id=arguments['--id'],
                      lenient=arguments['--lenient'],
                      disable_normalization=arguments['--disablenormalize'])

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import io
import six
import csv

class PrettyCsv(object):
    """Act as a csv DictWriter or console pretty-printer, according to whether outfile is set or not."""

    def __init__(self, outfile, fieldnames=None, maxwidth=120, **kwargs):
        self._fieldnames = fieldnames
        if outfile:
            self._rows = None
            if not outfile.endswith('.csv') and not outfile.endswith('.txt'):
                outfile += '.csv'
            if six.PY2:
                self._csvfile = open(outfile, 'wb')
            else:
                self._csvfile = io.TextIOWrapper(open(outfile, 'wb'), encoding='utf-8')
            self._csv = csv.writer(self._csvfile, **kwargs)
            if self._fieldnames is not None:
                # Write header
                self._csv.writerow(self._fieldnames)
        else:
            self._maxwidth = int(maxwidth)
            self._csv = None
            self._rows = []

    def writerow(self, row):
        if isinstance(row, dict):
            if self._fieldnames is None:
                raise Exception("You can't add a row as dictionnary w/o specifying fieldnames!")
            row = [ row.get(fieldname, None) for fieldname in self._fieldnames ]
        # Force to unicode
        row = [ six.u("") if v is None else v if isinstance(v, six.text_type) else six.u(str(v)) for v in row ]
        if self._csv:
            if six.PY2:
                self._csv.writerow([ v.encode('utf-8') for v in row ])
            else:
                self._csv.writerow(row)
        if self._rows is not None:
            self._rows.append(row)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if self._csv:
            self._csvfile.close()
        else:
            if self._fieldnames is not None:
                allrows = [ self._fieldnames ] + self._rows
            else:
                allrows = self._rows
            ncols = max(len(row) for row in allrows)
            colwidths = [ 0 for i in range(0, ncols) ]
            for row in allrows:
                for i in range(0, len(row)):
                    cell = row[i]
                    l = min(self._maxwidth, len(cell))
                    if cell and l > colwidths[i]:
                        colwidths[i] = l
            if self._fieldnames is not None:
                self._prettysep(colwidths)
                self._prettyprint(colwidths, self._fieldnames)
            self._prettysep(colwidths)
            for row in self._rows:
                self._prettyprint(colwidths, row)
            self._prettysep(colwidths)

    def _prettyprint(self, widths, row):
        s = "|"
        for width, cell in six.moves.zip_longest(widths, row, fillvalue=None):
            scell = six.u("") if cell is None else cell
            diff = width - len(scell)
            s += ' ' + (' ' * diff) + scell[:width] + ' |'
        print(s)

    def _prettysep(self, widths):
        s = "+"
        for width in widths:
            s += '-' * (width + 1) + '-+'
        print(s)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from _collections import defaultdict
from sqlalchemy.sql.expression import or_
from sqlalchemy.sql.functions import func
from sqlalchemy.orm import clear_mappers

from gtfslib.dao import Dao
from gtfslib.model import CalendarDate, Stop, StopTime, Route
from gtfslib.utils import fmttime

DUMMY_GTFS = "test/dummy.gtfs.zip"
DAO_URL = ""

class TestDemo(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    def test_demo(self):
        dao = Dao(DAO_URL, sql_logging=False)
        dao.load_gtfs(DUMMY_GTFS)

        print("List of stops named '...Bordeaux...':")
        stops_bordeaux = list(dao.stops(fltr=(Stop.stop_name.ilike('%Bordeaux%')) & (Stop.location_type == Stop.TYPE_STOP)))
        for stop in stops_bordeaux:
            print(stop.stop_name)

        print("List of routes passing by those stops:")
        routes_bordeaux = dao.routes(fltr=or_(StopTime.stop == stop for stop in stops_bordeaux))
        for route in routes_bordeaux:
            print("%s - %s" % (route.route_short_name, route.route_long_name))

        july4 = CalendarDate.ymd(2016, 7, 4)
        print("All departures from those stops on %s:" % (july4.as_date()))
        departures = list(dao.stoptimes(fltr=(or_(StopTime.stop == stop for stop in stops_bordeaux)) & (StopTime.departure_time != None) & (func.date(CalendarDate.date) == july4.date)))
        print("There is %d departures" % (len(departures)))
        for departure in departures:
            print("%30.30s %10.10s %-20.20s > %s" % (departure.stop.stop_name, fmttime(departure.departure_time), departure.trip.route.route_long_name, departure.trip.trip_headsign))

        print("Number of departures and time range per stop on %s:" % (july4.as_date()))
        departure_by_stop = defaultdict(list)
        for departure in departures:
            departure_by_stop[departure.stop].append(departure)
        for stop, deps in departure_by_stop.items():
            min_dep = min(d.departure_time for d in deps)
            max_dep = max(d.departure_time for d in deps)
            print("%30.30s %3d departures (from %s to %s)" % (stop.stop_name, len(deps), fmttime(min_dep), fmttime(max_dep)))

        # Compute the average distance and time to next stop by route type
        ntd = [ [0, 0, 0.0] for type in range(0, Route.TYPE_FUNICULAR + 1) ]
        for departure in departures:
            # The following is guaranteed to succeed as we have departure_time == Null for last stop time in trip
            next_arrival = departure.trip.stop_times[departure.stop_sequence + 1]
            hop_dist = next_arrival.shape_dist_traveled - departure.shape_dist_traveled
            hop_time = next_arrival.arrival_time - departure.departure_time
            route_type = departure.trip.route.route_type
            ntd[route_type][0] += 1
            ntd[route_type][1] += hop_time
            ntd[route_type][2] += hop_dist
        for route_type in range(0, len(ntd)):
            n, t, d = ntd[route_type]
            if n > 0:
                print("The average distance to the next stop on those departures for route type %d is %.2f meters" % (route_type, d / n))
                print("The average time in sec to the next stop on those departures for route type %d is %s" % (route_type, fmttime(t / n)))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""
import logging

from gtfslib.model import Agency, FeedInfo, Route, Calendar, CalendarDate, Stop, \
    Trip, StopTime, Transfer, Shape, ShapePoint, Zone, FareAttribute, FareRule
from gtfslib.spatial import DistanceCache, orthodromic_distance,\
    orthodromic_seg_distance
from gtfslib.utils import timing, fmttime, ContinousPiecewiseLinearFunc

logger = logging.getLogger('libgtfs')

DOW_NAMES = { 0: 'monday', 1: 'tuesday', 2: 'wednesday', 3: 'thursday', 4: 'friday', 5: 'saturday', 6: 'sunday' }

def _toint(s, default_value=None):
    if s is None or len(s) == 0:
        if default_value is None:
            raise ValueError()
        return default_value
    return int(s)

def _timetoint(s, default_value=None):
    if s is None or len(s) == 0:
        if default_value is None:
            raise ValueError()
        return default_value
    hms = s.split(':')
    if len(hms) != 3:
        raise ValueError("Invalid date/time: %s" % s)
    return int(hms[0]) * 3600 + int(hms[1]) * 60 + int(hms[2])

def _tofloat(s, default_value=None):
    if s is None or len(s) == 0:
        return default_value
    return float(s)

def _todate(s, default_value=None):
    if s is None or len(s) == 0:
        return default_value
    return CalendarDate.fromYYYYMMDD(s).as_date()

class _CacheEntry(object):

    def __init__(self, distance):
        self._distance = distance
        self._next_entries = {}

    def next_entry(self, stop):
        return self._next_entries.get(stop.stop_id)

    def distance(self):
        return self._distance

    def insert(self, stop, distance):
        next_entry = _CacheEntry(distance)
        self._next_entries[stop.stop_id] = next_entry
        return next_entry

class _OdometerShape(object):

    # A 0.1% cone - Coefficient to defavor points further away on a shape
    # when doing stop snapping to shape. The summit angle of the offset cone
    # is 2 * arctan(K).
    K = 0.001

    def __init__(self, shape):
        self._shape = shape
        self._cache = _CacheEntry(None)
        self._cache_hit = 0
        self._cache_miss = 0
        if all(pt.shape_dist_traveled != -999999 for pt in shape.points):
            self._xdist = ContinousPiecewiseLinearFunc()
        else:
            self._xdist = None
        # Normalize the shape here:
        # 1) dist_traveled to meters
        # 2) pt_seq to contiguous numbering from 0
        ptseq = 0
        distance_meters = 0.0
        last_pt = None
        shape.points.sort(key=lambda p: p.shape_pt_sequence)
        for pt in shape.points:
            if last_pt is not None:
                # Note: we do not use distance cache, as most probably
                # many of the points will be different from each other.
                distance_meters += orthodromic_distance(last_pt, pt)
            last_pt = pt
            pt.shape_pt_sequence = ptseq
            old_distance = pt.shape_dist_traveled
            pt.shape_dist_traveled = distance_meters
            # Remember the distance mapping for stop times
            if self._xdist:
                self._xdist.append(old_distance, pt.shape_dist_traveled)
            ptseq += 1

    def reset(self):
        self._distance = 0
        self._istart = 0
        self._cache_cursor = self._cache

    def dist_traveled(self, stop, old_dist_traveled):
        if old_dist_traveled and self._xdist:
            # Case 1: we have in the original data shape_dist_traveled
            # We need to remap from the old scale to the new meter scale
            return self._xdist.interpolate(old_dist_traveled)
        else:
            # Case 2: we do not have original shape_dist_traveled
            # We need to determine ourselves where in the shape we lie
            # TODO Implement a cache, this can be slow for lots of trips
            # and the result is the same for the same pattern
            # Check the cache first
            cache_entry = self._cache_cursor.next_entry(stop)
            if cache_entry is not None:
                self._cache_cursor = cache_entry
                self._cache_hit += 1
                return cache_entry.distance()
            min_dist = 1e20
            best_i = self._istart
            best_dist = 0
            for i in range(self._istart, len(self._shape.points) - 1):
                a = self._shape.points[i]
                b = self._shape.points[i+1]
                dist, pdist = orthodromic_seg_distance(stop, a, b)
                newdist = a.shape_dist_traveled + pdist
                howfar = newdist - self._distance
                # Add a slight "cone" offset. There are pathological
                # cases with backtracking shapes where the best distance
                # is slightly better way further (for eg 0.01m) than at
                # the starting point (for eg 0.02m). In that case we should
                # obviously keep the first point instead of moving too fast
                # to the shape end. That offset should help for some cases.
                dist += howfar * self.K
                if dist < min_dist:
                    min_dist = dist
                    best_i = i
                    best_dist = newdist
            if best_dist > self._distance:
                self._distance = best_dist
            else:
                delta = self._distance - best_dist
                if delta > 10:
                    # This is harmless if the backtracking distance is small.
                    # We have lots of false positive (<<1m) due to rounding errors.
                    logger.warn("Backtracking of %f m detected in shape %s for stop %s (%s) (%f,%f) at distance %f < %f m on segment #[%d-%d]" % (
                                delta, self._shape.shape_id, stop.stop_id, stop.stop_name, stop.stop_lat, stop.stop_lon, best_dist, self._distance, best_i, best_i+1))
            self._istart = best_i
            self._cache_miss += 1
            self._cache_cursor = self._cache_cursor.insert(stop, self._distance)
            return self._distance

    def _debug_cache(self):
        logger.debug("Shape %s: Cache hit: %d, misses: %d" % (self._shape.shape_id, self._cache_hit, self._cache_miss))

class _Odometer(object):
    _odoshp = None
    _dcache = None

    def normalize_and_register_shape(self, shape):
        self._odoshp = _OdometerShape(shape)
        self.reset()

    def register_noshape(self):
        self._odoshp = None
        self.reset()

    def reset(self):
        if self._odoshp is not None:
            self._odoshp.reset()
        self._distance = 0
        self._last_stop = None
        self._dcache = DistanceCache()

    def dist_traveled(self, stop, old_dist_traveled):
        if self._odoshp is not None:
            # We have a shape, use it
            return self._odoshp.dist_traveled(stop, old_dist_traveled)
        # We do not have shape, use straight-line distance between
        # consecutive stops
        if self._last_stop is not None:
            self._distance += self._dcache.orthodromic_distance(self._last_stop, stop)
        self._last_stop = stop
        return self._distance

    def _debug_cache(self):
        self._odoshp._debug_cache()

@timing
def _convert_gtfs_model(feed_id, gtfs, dao, lenient=False, disable_normalization=False):
    
    feedinfo2 = None
    logger.info("Importing feed ID '%s'" % feed_id)
    n_feedinfo = 0
    for feedinfo in gtfs.feedinfo():
        n_feedinfo += 1
        if n_feedinfo > 1:
            logger.error("Feed info should be unique if defined. Taking first one." % (n_feedinfo))
            break
        # TODO Automatically compute from calendar range if missing?
        feedinfo['feed_start_date'] = _todate(feedinfo.get('feed_start_date'))
        feedinfo['feed_end_date'] = _todate(feedinfo.get('feed_end_date'))
        feedinfo2 = FeedInfo(feed_id, **feedinfo)
    if feedinfo2 is None:
        # Optional, generate empty feed info
        feedinfo2 = FeedInfo(feed_id)
    dao.add(feedinfo2)
    dao.flush()
    logger.info("Imported %d feedinfo" % n_feedinfo)

    logger.info("Importing agencies...")
    n_agencies = 0
    single_agency = None
    agency_ids = set()
    for agency in gtfs.agencies():
        # agency_id is optional only if we have a single agency
        if n_agencies == 0 and agency.get('agency_id') is None:
            agency['agency_id'] = ''
        agency2 = Agency(feed_id, **agency)
        if n_agencies == 0:
            single_agency = agency2
        else:
            single_agency = None
        n_agencies += 1
        dao.add(agency2)
        agency_ids.add(agency2.agency_id)
    dao.flush()
    logger.info("Imported %d agencies" % n_agencies)

    def import_stop(stop, stoptype, zone_ids, item_ids, station_ids=None):
        zone_id = stop.get('zone_id')
        if zone_id and zone_id not in zone_ids:
            # Lazy-creation of zone
            zone = Zone(feed_id, zone_id)
            zone_ids.add(zone_id)
            dao.add(zone)
        stop['location_type'] = _toint(stop.get('location_type'), Stop.TYPE_STOP)
        if stop['location_type'] != stoptype:
            return 0
        stop['wheelchair_boarding'] = _toint(stop.get('wheelchair_boarding'), Stop.WHEELCHAIR_UNKNOWN)
        lat = _tofloat(stop.get('stop_lat'), None)
        lon = _tofloat(stop.get('stop_lon'), None)
        if lat is None or lon is None:
            if lenient:
                logger.error("Missing lat/lon for '%s', set to default (0,0)" % (stop,))
                if lat is None:
                    lat = 0
                if lon is None:
                    lon = 0
            else:
                raise ValueError("Missing mandatory lat/lon for '%s'." % (stop,))
        stop['stop_lat'] = lat
        stop['stop_lon'] = lon
        # This field has been renamed for consistency
        parent_id = stop.get('parent_station')
        stop['parent_station_id'] = parent_id if parent_id else None
        if parent_id and station_ids and parent_id not in station_ids:
            if lenient:
                logger.error("Parent station ID '%s' in '%s' is invalid, resetting." % (parent_id, stop))
                stop['parent_station_id'] = None
            else:
                raise KeyError("Parent station ID '%s' in '%s' is invalid." % (parent_id, stop))
        stop.pop('parent_station', None)
        stop2 = Stop(feed_id, **stop)
        dao.add(stop2)
        item_ids.add(stop2.stop_id)
        return 1

    stop_ids = set()
    station_ids = set()
    zone_ids = set()
    logger.info("Importing zones, stations and stops...")
    n_stations = n_stops = 0
    for station in gtfs.stops():
        n_stations += import_stop(station, Stop.TYPE_STATION, zone_ids, station_ids)
    for stop in gtfs.stops():
        n_stops += import_stop(stop, Stop.TYPE_STOP, zone_ids, stop_ids, station_ids)
    dao.flush()
    logger.info("Imported %d zones, %d stations and %d stops" % (len(zone_ids), n_stations, n_stops))

    logger.info("Importing transfers...")
    n_transfers = 0
    for transfer in gtfs.transfers():
        from_stop_id = transfer.get('from_stop_id')
        to_stop_id = transfer.get('to_stop_id')
        transfer['transfer_type'] = _toint(transfer.get('transfer_type'), 0)
        for stop_id in (from_stop_id, to_stop_id):
            if stop_id not in station_ids and stop_id not in stop_ids:
                if lenient:
                    logger.error("Stop ID '%s' in '%s' is invalid, skipping." % (stop_id, transfer))
                    continue
                else:
                    raise KeyError("Stop ID '%s' in '%s' is invalid." % (stop_id, transfer))
        transfer2 = Transfer(feed_id, **transfer)
        n_transfers += 1
        dao.add(transfer2)
    dao.flush()
    logger.info("Imported %d transfers" % (n_transfers))
    
    logger.info("Importing routes...")
    n_routes = 0
    route_ids = set()
    for route in gtfs.routes():
        route['route_type'] = int(route.get('route_type'))
        agency_id = route.get('agency_id')
        if (agency_id is None or len(agency_id) == 0) and single_agency is not None:
            # Route.agency is optional if only a single agency exists.
            agency_id = route['agency_id'] = single_agency.agency_id
        if agency_id not in agency_ids:
            if lenient:
                logger.error("Agency ID '%s' in '%s' is invalid, skipping route." % (agency_id, route))
                continue
            else:
                raise KeyError("agency ID '%s' in '%s' is invalid." % (agency_id, route))
        route2 = Route(feed_id, **route)
        dao.add(route2)
        route_ids.add(route2.route_id)
        n_routes += 1
    dao.flush()
    logger.info("Imported %d routes" % n_routes)

    logger.info("Importing fares...")
    n_fares = 0
    for fare_attr in gtfs.fare_attributes():
        fare_id = fare_attr.get('fare_id')
        fare_price = _tofloat(fare_attr.get('price'))
        currency_type = fare_attr.get('currency_type')
        payment_method = _toint(fare_attr.get('payment_method'))
        n_transfers = None
        if fare_attr.get('transfers') is not None:
            n_transfers = _toint(fare_attr.get('transfers'))
        transfer_duration = None
        if fare_attr.get('transfer_duration') is not None:
            transfer_duration = _toint(fare_attr.get('transfer_duration'))
        fare = FareAttribute(feed_id, fare_id, fare_price, currency_type,
                             payment_method, n_transfers, transfer_duration)
        dao.add(fare)
        n_fares += 1
    dao.flush()
    fare_rules = set()
    for fare_rule in gtfs.fare_rules():
        fare_rule2 = FareRule(feed_id, **fare_rule)
        if fare_rule2 in fare_rules:
            if lenient:
                logger.error("Duplicated fare rule (%s), skipping." % (fare_rule2))
                continue
            else:
                raise KeyError("Duplicated fare rule (%s)" % (fare_rule2))
        dao.add(fare_rule2)
        fare_rules.add(fare_rule2)
    dao.flush()
    logger.info("Imported %d fare and %d rules" % (n_fares, len(fare_rules)))

    logger.info("Importing calendars...")
    calanddates2 = {}
    for calendar in gtfs.calendars():
        calid = calendar.get('service_id')
        calendar2 = Calendar(feed_id, calid)
        dates2 = []
        start_date = CalendarDate.fromYYYYMMDD(calendar.get('start_date'))
        end_date = CalendarDate.fromYYYYMMDD(calendar.get('end_date'))
        for d in CalendarDate.range(start_date, end_date.next_day()):
            if int(calendar.get(DOW_NAMES[d.dow()])):
                dates2.append(d)
        calanddates2[calid] = (calendar2, set(dates2))

    logger.info("Normalizing calendar dates...")
    for caldate in gtfs.calendar_dates():
        calid = caldate.get('service_id')
        date2 = CalendarDate.fromYYYYMMDD(caldate.get('date'))
        addremove = int(caldate.get('exception_type'))
        if calid in calanddates2:
            calendar2, dates2 = calanddates2[calid]
        else:
            calendar2 = Calendar(feed_id, calid)
            dates2 = set([])
            calanddates2[calid] = (calendar2, dates2)
        if addremove == 1:
            dates2.add(date2)
        elif addremove == 2:
            if date2 in dates2:
                dates2.remove(date2)
    n_calendars = 0
    n_caldates = 0
    calendar_ids = set()
    for (calendar2, dates2) in calanddates2.values():
        calendar2.dates = [ d for d in dates2 ]
        dao.add(calendar2)
        calendar_ids.add(calendar2.service_id)
        n_calendars += 1
        n_caldates += len(calendar2.dates)
    dao.flush()
    logger.info("Imported %d calendars and %d dates" % (n_calendars, n_caldates))

    logger.info("Importing shapes...")
    n_shape_pts = 0
    shape_ids = set()
    shapepts_q = []
    for shpt in gtfs.shapes():
        shape_id = shpt.get('shape_id')
        if shape_id not in shape_ids:
            dao.add(Shape(feed_id, shape_id))
            dao.flush()
            shape_ids.add(shape_id)
        pt_seq = _toint(shpt.get('shape_pt_sequence'))
        # This field is optional
        dist_traveled = _tofloat(shpt.get('shape_dist_traveled'), -999999)
        lat = _tofloat(shpt.get('shape_pt_lat'))
        lon = _tofloat(shpt.get('shape_pt_lon'))
        shape_point = ShapePoint(feed_id, shape_id, pt_seq, lat, lon, dist_traveled)
        shapepts_q.append(shape_point)
        n_shape_pts += 1
        if n_shape_pts % 100000 == 0:
            logger.info("%d shape points" % n_shape_pts)
            dao.bulk_save_objects(shapepts_q)
            dao.flush()
            shapepts_q = []
    dao.bulk_save_objects(shapepts_q)
    dao.flush()
    logger.info("Imported %d shapes and %d points" % (len(shape_ids), n_shape_pts))

    logger.info("Importing trips...")
    n_trips = 0
    trips_q = []
    trip_ids = set()
    for trip in gtfs.trips():
        trip['wheelchair_accessible'] = _toint(trip.get('wheelchair_accessible'), Trip.WHEELCHAIR_UNKNOWN)
        trip['bikes_allowed'] = _toint(trip.get('bikes_allowed'), Trip.BIKES_UNKNOWN)
        cal_id = trip.get('service_id')
        if cal_id not in calendar_ids:
            if lenient:
                logger.error("Calendar ID '%s' in '%s' is invalid. Skipping trip." % (cal_id, trip))
                continue
            else:
                raise KeyError("Calendar ID '%s' in '%s' is invalid." % (cal_id, trip))
        route_id = trip.get('route_id')
        if route_id not in route_ids:
            if lenient:
                logger.error("Route ID '%s' in '%s' is invalid. Skipping trip." % (route_id, trip))
                continue
            else:
                raise KeyError("Route ID '%s' in trip '%s' is invalid." % (route_id, trip))
        trip2 = Trip(feed_id, frequency_generated=False, **trip)
        
        trips_q.append(trip2)
        n_trips += 1
        if n_trips % 10000 == 0:
            dao.bulk_save_objects(trips_q)
            dao.flush()
            logger.info('%s trips' % n_trips)
            trips_q = []

        trip_ids.add(trip.get('trip_id'))
    dao.bulk_save_objects(trips_q)
    dao.flush()
    
    logger.info("Imported %d trips" % n_trips)

    logger.info("Importing stop times...")
    n_stoptimes = 0
    stoptimes_q = []
    for stoptime in gtfs.stop_times():
        stopseq = _toint(stoptime.get('stop_sequence'))
        # Mark times to interpolate later on 
        arrtime = _timetoint(stoptime.get('arrival_time'), -999999)
        deptime = _timetoint(stoptime.get('departure_time'), -999999)
        if arrtime == -999999:
            arrtime = deptime
        if deptime == -999999:
            deptime = arrtime
        interp = arrtime < 0 and deptime < 0
        shpdist = _tofloat(stoptime.get('shape_dist_traveled'), -999999)
        pkptype = _toint(stoptime.get('pickup_type'), StopTime.PICKUP_DROPOFF_REGULAR)
        drptype = _toint(stoptime.get('drop_off_type'), StopTime.PICKUP_DROPOFF_REGULAR)
        trip_id = stoptime.get('trip_id')
        if trip_id not in trip_ids:
            if lenient:
                logger.error("Trip ID '%s' in '%s' is invalid. Skipping stop time." % (trip_id, stoptime))
                continue
            else:
                raise KeyError("Trip ID '%s' in '%s' is invalid." % (trip_id, stoptime))
        stop_id = stoptime.get('stop_id')
        if stop_id not in stop_ids:
            if lenient:
                logger.error("Stop ID '%s' in '%s' is invalid. Skipping stop time." % (stop_id, stoptime))
                continue
            else:
                raise KeyError("Trip ID '%s' in stoptime '%s' is invalid." % (stop_id, stoptime))
        stoptime2 = StopTime(feed_id, trip_id, stop_id,
                stop_sequence=stopseq, arrival_time=arrtime, departure_time=deptime,
                shape_dist_traveled=shpdist, interpolated=interp,
                pickup_type=pkptype, drop_off_type=drptype,
                stop_headsign=stoptime.get('stop_headsign'))
        stoptimes_q.append(stoptime2)
        n_stoptimes += 1
        # Commit every now and then
        if n_stoptimes % 50000 == 0:
            logger.info("%d stop times" % n_stoptimes)
            dao.bulk_save_objects(stoptimes_q)
            dao.flush()
            stoptimes_q = []
    dao.bulk_save_objects(stoptimes_q)

    logger.info("Imported %d stop times" % n_stoptimes)
    logger.info("Committing")
    dao.flush()
    # TODO Add option to enable/disable this commit
    # to ensure import is transactionnal
    dao.commit()
    logger.info("Commit done")

    def normalize_trip(trip, odometer):
        stopseq = 0
        n_stoptimes = len(trip.stop_times)
        last_stoptime_with_time = None
        to_interpolate = []
        odometer.reset()
        for stoptime in trip.stop_times:
            stoptime.stop_sequence = stopseq
            stoptime.shape_dist_traveled = odometer.dist_traveled(stoptime.stop,
                        stoptime.shape_dist_traveled if stoptime.shape_dist_traveled != -999999 else None)
            if stopseq == 0:
                # Force first arrival time to NULL
                stoptime.arrival_time = None
            if stopseq == n_stoptimes - 1:
                # Force last departure time to NULL
                stoptime.departure_time = None
            if stoptime.interpolated:
                to_interpolate.append(stoptime)
            else:
                if len(to_interpolate) > 0:
                    # Interpolate
                    if last_stoptime_with_time is None:
                        logger.error("Cannot interpolate missing time at trip start: %s" % trip)
                        for stti in to_interpolate:
                            # Use first defined time as fallback value.
                            stti.arrival_time = stoptime.arrival_time
                            stti.departure_time = stoptime.arrival_time
                    else:
                        tdist = stoptime.shape_dist_traveled - last_stoptime_with_time.shape_dist_traveled
                        ttime = stoptime.arrival_time - last_stoptime_with_time.departure_time
                        for stti in to_interpolate:
                            fdist = stti.shape_dist_traveled - last_stoptime_with_time.shape_dist_traveled
                            t = last_stoptime_with_time.departure_time + ttime * fdist // tdist
                            stti.arrival_time = t
                            stti.departure_time = t
                to_interpolate = []
                last_stoptime_with_time = stoptime
            stopseq += 1
        if len(to_interpolate) > 0:
            # Should not happen, but handle the case, we never know
            if last_stoptime_with_time is None:
                logger.error("Cannot interpolate missing time, no time at all: %s" % trip)
                # Keep times NULL (TODO: or remove the trip?)
            else:
                logger.error("Cannot interpolate missing time at trip end: %s" % trip)
                for stti in to_interpolate:
                    # Use last defined time as fallback value
                    stti.arrival_time = last_stoptime_with_time.departure_time
                    stti.departure_time = last_stoptime_with_time.departure_time

    if disable_normalization:
        logger.info("Skipping shapes and trips normalization")
    else:
        logger.info("Normalizing shapes and trips...")
        nshapes = 0
        ntrips = 0
        odometer = _Odometer()
        # Process shapes and associated trips
        for shape in dao.shapes(fltr=Shape.feed_id == feed_id, prefetch_points=True, batch_size=50):
            # Shape will be registered in the normalize
            odometer.normalize_and_register_shape(shape)
            for trip in dao.trips(fltr=(Trip.feed_id == feed_id) & (Trip.shape_id == shape.shape_id), prefetch_stop_times=True, prefetch_stops=True, batch_size=800):
                normalize_trip(trip, odometer)
                ntrips += 1
                if ntrips % 1000 == 0:
                    logger.info("%d trips, %d shapes" % (ntrips, nshapes))
                    dao.flush()
            nshapes += 1
            #odometer._debug_cache()
        # Process trips w/o shapes
        for trip in dao.trips(fltr=(Trip.feed_id == feed_id) & (Trip.shape_id == None), prefetch_stop_times=True, prefetch_stops=True, batch_size=800):
            odometer.register_noshape()
            normalize_trip(trip, odometer)
            ntrips += 1
            if ntrips % 1000 == 0:
                logger.info("%d trips" % ntrips)
                dao.flush()
        dao.flush()
        logger.info("Normalized %d trips and %d shapes" % (ntrips, nshapes))

    # Note: we expand frequencies *after* normalization
    # for performances purpose only: that minimize the
    # number of trips to normalize. We can do that since
    # the expansion is neutral trip-normalization-wise.
    logger.info("Expanding frequencies...")
    n_freq = 0
    n_exp_trips = 0
    trips_to_delete = []
    for frequency in gtfs.frequencies():
        trip_id = frequency.get('trip_id')
        if trip_id not in trip_ids:
            if lenient:
                logger.error("Trip ID '%s' in '%s' is invalid. Skipping frequency." % (trip_id, frequency))
                continue
            else:
                raise KeyError("Trip ID '%s' in '%s' is invalid." % (trip_id, frequency))
        trip = dao.trip(trip_id, feed_id=feed_id)
        start_time = _timetoint(frequency.get('start_time'))
        end_time = _timetoint(frequency.get('end_time'))
        headway_secs = _toint(frequency.get('headway_secs'))
        exact_times = _toint(frequency.get('exact_times'), Trip.TIME_APPROX)
        for trip_dep_time in range(start_time, end_time, headway_secs):
            # Here we assume departure time are all different.
            # That's a requirement in the GTFS specs, but this may break.
            # TODO Make the expanded trip ID generation parametrable.
            trip_id2 = trip.trip_id + "@" + fmttime(trip_dep_time)
            trip2 = Trip(feed_id, trip_id2, trip.route_id, trip.service_id,
                         wheelchair_accessible=trip.wheelchair_accessible,
                         bikes_allowed=trip.bikes_allowed,
                         exact_times=exact_times,
                         frequency_generated=True,
                         trip_headsign=trip.trip_headsign,
                         trip_short_name=trip.trip_short_name,
                         direction_id=trip.direction_id,
                         block_id=trip.block_id)
            trip2.stop_times = []
            base_time = trip.stop_times[0].departure_time
            for stoptime in trip.stop_times:
                arrtime = None if stoptime.arrival_time is None else stoptime.arrival_time - base_time + trip_dep_time
                deptime = None if stoptime.departure_time is None else stoptime.departure_time - base_time + trip_dep_time
                stoptime2 = StopTime(feed_id, trip_id2, stoptime.stop_id, stoptime.stop_sequence,
                            arrival_time=arrtime,
                            departure_time=deptime,
                            shape_dist_traveled=stoptime.shape_dist_traveled,
                            interpolated=stoptime.interpolated,
                            timepoint=stoptime.timepoint,
                            pickup_type=stoptime.pickup_type,
                            drop_off_type=stoptime.drop_off_type)
                trip2.stop_times.append(stoptime2)
            n_exp_trips += 1
            # This will add the associated stop times
            dao.add(trip2)
        # Do not delete trip now, as two frequency can refer to same trip
        trips_to_delete.append(trip)
        n_freq += 1
    for trip in trips_to_delete:
        # This also delete the associated stop times
        dao.delete(trip)
    dao.flush()
    dao.commit()
    logger.info("Expanded %d frequencies to %d trips." % (n_freq, n_exp_trips))

    logger.info("Feed '%s': import done." % feed_id)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from inspect import isclass

import sqlalchemy
from sqlalchemy.orm import subqueryload
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.orm.util import aliased
from sqlalchemy.exc import InvalidRequestError

from gtfslib.converter import _convert_gtfs_model
from gtfslib.csvgtfs import Gtfs, ZipFileSource
from gtfslib.model import FeedInfo, Agency, Route, Calendar, CalendarDate, Stop, \
    Trip, StopTime, Transfer, Shape, Zone, FareAttribute, FareRule, ShapePoint
from gtfslib.orm import _Orm
from gtfslib.utils import group_pairs

class Dao(object):
    """
    Note to developer: Please do not use fancy magic, such as automatically generating methods,
    as this may break auto-complete and thus make the use of this class more difficult.
    """

    def __init__(self, db="", sql_logging=False, schema=None):
        if db == "" or db is None:
            # In-memory SQLite
            connect_url = "sqlite:///"
        if "://" in db:
            # User has provided a full path
            connect_url = db
        else:
            # Assume a SQLite file
            connect_url = "sqlite:///%s" % db
        engine = sqlalchemy.create_engine(connect_url, echo=sql_logging)
        self._orm = _Orm(engine, schema=schema)
        Session = sessionmaker(bind=engine)
        self._session = Session()
        self._stoptime1 = aliased(StopTime, name="first_stop_time")
        self._stoptime2 = aliased(StopTime, name="second_stop_time")
        self._transfer_fromstop = aliased(Stop, name="tr_from_stop")
        self._transfer_tostop = aliased(Stop, name="tr_to_stop")

    def session(self):
        return self._session

    def bulk_save_objects(self, objects):
        return self._session.bulk_save_objects(objects)
        
    def add(self, obj):
        self._session.add(obj)
        
    def add_all(self, objs):
        self._session.add_all(objs)

    def delete(self, obj):
        self._session.delete(obj)
        
    def delete_feed(self, feed_id):
        self._session.query(FareRule).filter(FareRule.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(FareAttribute).filter(FareAttribute.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(StopTime).filter(StopTime.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Trip).filter(Trip.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(ShapePoint).filter(ShapePoint.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Shape).filter(Shape.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(CalendarDate).filter(CalendarDate.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Calendar).filter(Calendar.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Route).filter(Route.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Stop).filter(Stop.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Agency).filter(Agency.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(Zone).filter(Zone.feed_id == feed_id).delete(synchronize_session=False)
        self._session.query(FeedInfo).filter(FeedInfo.feed_id == feed_id).delete()

    def commit(self):
        self._session.commit()
        
    def flush(self):
        self._session.flush()

    def feed(self, feed_id=""):
        return self._session.query(FeedInfo).get(feed_id)

    def feeds(self):
        return self._session.query(FeedInfo).all()

    def agency(self, agency_id, feed_id="", prefetch_routes=False):
        query = self._session.query(Agency)
        if prefetch_routes:
            query = query.options(subqueryload('routes'))
        return query.get((feed_id, agency_id))
        
    def agencies(self, fltr=None, prefetch_routes=False):
        query = self._session.query(Agency).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_routes:
            query = query.options(subqueryload('routes'))
        return query.all()

    def zone(self, zone_id, feed_id="", prefetch_stops=False):
        query = self._session.query(Zone)
        if prefetch_stops:
            query = query.options(subqueryload('stops'))
        return query.get((feed_id, zone_id))

    def zones(self, fltr=None, prefetch_stops=False):
        query = self._session.query(Zone).distinct()
        if fltr is not None:
            query = query.filter(fltr)
        if prefetch_stops:
            query = query.options(subqueryload('stops'))
        return query.all()

    def stop(self, stop_id, feed_id="", prefetch_parent=True, prefetch_substops=True):
        query = self._session.query(Stop)
        if prefetch_parent:
            query = query.options(subqueryload('parent_station'))
        if prefetch_substops:
            query = query.options(subqueryload('sub_stops'))
        return query.get((feed_id, stop_id))
    
    def stops(self, fltr=None, prefetch_parent=True, prefetch_substops=True, batch_size=1000):
        idquery = self._session.query(Stop.feed_id, Stop.stop_id).distinct()
        if fltr is not None:
            idquery = _AutoJoiner(self._orm, idquery, fltr).autojoin()
            idquery = idquery.filter(fltr)
        # Only query IDs first
        stopids = idquery.all()
        def query_factory():
            query = self._session.query(Stop)
            if prefetch_parent:
                query = query.options(subqueryload('parent_station'))
                if prefetch_substops:
                    query = query.options(subqueryload('sub_stops'))
            return query
        return self._page_query(query_factory, Stop.feed_id, Stop.stop_id, stopids, batch_size)

    def in_area(self, area):
        """Return a filter filtering stops in the given area (RectangularArea...)"""
        return (Stop.stop_lat >= area.min_lat) & (Stop.stop_lat <= area.max_lat) & (Stop.stop_lon >= area.min_lon) & (Stop.stop_lon <= area.max_lon)

    def in_bounds(self, min_lat, min_lon, max_lat, max_lon):
        """Return a filter filtering stops in the given bounds."""
        return (Stop.stop_lat >= min_lat) & (Stop.stop_lat <= max_lat) & (Stop.stop_lon >= min_lon) & (Stop.stop_lon <= max_lon)

    def transfer(self, from_stop_id, to_stop_id, feed_id="", prefetch_stops=True):
        query = self._session.query(Transfer)
        if prefetch_stops:
            query = query.options(subqueryload('from_stop'), subqueryload('to_stop'))
        return query.get((feed_id, from_stop_id, to_stop_id))

    def transfer_from_stop(self):
        return self._transfer_fromstop
    
    def transfer_to_stop(self):
        return self._transfer_tostop

    def transfers(self, fltr=None, prefetch_stops=True):
        query = self._session.query(Transfer).distinct()
        if fltr is not None:
            query = query.join(self._transfer_fromstop, 'from_stop')
            query = query.join(self._transfer_tostop, 'to_stop')
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_stops:
            query = query.options(subqueryload('from_stop'), subqueryload('to_stop'))
        return query.all()

    def route(self, route_id, feed_id=""):
        return self._session.query(Route).get((feed_id, route_id))

    def routes(self, fltr=None, prefetch_trips=False):
        query = self._session.query(Route).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_trips:
            query = query.options(subqueryload('trips'))
        return query.all()
    
    def calendar(self, service_id, feed_id="", prefetch_dates=True, prefetch_trips=False, prefetch_stop_times=False):
        query = self._session.query(Calendar)
        if prefetch_stop_times:
            prefetch_trips = True
        if prefetch_trips:
            loadopt = subqueryload('trips')
            if prefetch_stop_times:
                loadopt = loadopt.subqueryload('stop_times')
            query = query.options(loadopt)
        if prefetch_dates:
            query = query.options(subqueryload('dates'))
        return query.get((feed_id, service_id))
    
    def calendars(self, fltr=None, prefetch_dates=True, prefetch_trips=False):
        query = self._session.query(Calendar).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_dates:
            query = query.options(subqueryload('dates'))
        if prefetch_trips:
            query = query.options(subqueryload('trips'))
        return query.all()
    
    def calendar_dates(self, fltr=None, prefetch_calendars=True, prefetch_trips=False):
        """Load calendar dates object. This may be not what you need, as it will return
           lots of duplicated dates (ie one date per calendar). If you only need distinct
           dates from a query (for example the set of date where a stop is active), you
           can use the calendar_dates_date() function instead."""
        query = self._session.query(CalendarDate).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_calendars:
            query = query.options(subqueryload('calendar'))
        if prefetch_trips:
            query = query.options(subqueryload('calendar.trips'))
        return query.all()

    def calendar_dates_date(self, fltr=None):
        """Return a set of distinct dates (python date). It will properly merge
           identical dates from different calendars, thus returning a proper
           set of distinct dates. In doubt with calendar_dates, this method is
           in all probability the one you want/need."""
        query = self._session.query(CalendarDate.date).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        return [ date for (date,) in query.all() ]

    def trip(self, trip_id, feed_id="", prefetch_stop_times=True):
        query = self._session.query(Trip)
        if prefetch_stop_times:
            query = query.options(subqueryload('stop_times'))
        return query.get((feed_id, trip_id))
    
    def trips(self, fltr=None, prefetch_stop_times=True, prefetch_routes=False, prefetch_stops=False, prefetch_calendars=False, batch_size=800):
        idquery = self._session.query(Trip.feed_id, Trip.trip_id).distinct()
        if fltr is not None:
            idquery = _AutoJoiner(self._orm, idquery, fltr).autojoin()
            idquery = idquery.filter(fltr)
        # Only query IDs first
        tripids = idquery.all()
        def query_factory():
            query = self._session.query(Trip)
            _prefetch_stop_times = prefetch_stop_times
            if prefetch_stops:
                _prefetch_stop_times = True
            if _prefetch_stop_times:
                loadopt = subqueryload('stop_times')
                if prefetch_stops:
                    loadopt = loadopt.subqueryload('stop')
                query = query.options(loadopt)
            if prefetch_routes:
                query = query.options(subqueryload('route'))
            if prefetch_calendars:
                query = query.options(subqueryload('calendar'))
            return query
        return self._page_query(query_factory, Trip.feed_id, Trip.trip_id, tripids, batch_size)

    def stoptimes(self, fltr=None, prefetch_trips=True, prefetch_stop_times=False):
        query = self._session.query(StopTime).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_stop_times:
            prefetch_trips = True
        if prefetch_trips:
            loadopt = subqueryload('trip')
            if prefetch_stop_times:
                loadopt = loadopt.subqueryload('stop_times')
            query = query.options(loadopt)
        # Note: ID batching would be difficult to implement for StopTime
        # as StopTime do have a composite-primary composed of 3 elements
        # and 2 of them (trip_id + stop_seq) can't be grouped easily.
        return query.all()

    def hop_first(self):
        return self._stoptime1

    def hop_second(self):
        return self._stoptime2

    def hops(self, delta=1, fltr=None, prefetch_trips=True, prefetch_stop_times=False):
        query = self._session.query(self._stoptime1, self._stoptime2).filter((self._stoptime1.trip_id == self._stoptime2.trip_id) & ((self._stoptime1.stop_sequence + delta) == self._stoptime2.stop_sequence)).distinct()
        if fltr is not None:
            query = _AutoJoiner(self._orm, query, fltr).autojoin()
            query = query.filter(fltr)
        if prefetch_stop_times:
            prefetch_trips = True
        if prefetch_trips:
            loadopt = subqueryload('trip')
            if prefetch_stop_times:
                loadopt = loadopt.subqueryload('stop_times')
            query = query.options(loadopt)
        return query.all()

    def shape(self, shape_id, feed_id="", prefetch_shape_points=True):
        query = self._session.query(Shape)
        if prefetch_shape_points:
            query = query.options(subqueryload('points'))
        return query.get((feed_id, shape_id))

    def shapes(self, fltr=None, prefetch_points=True, batch_size=100):
        idquery = self._session.query(Shape.feed_id, Shape.shape_id).distinct()
        if fltr is not None:
            idquery = _AutoJoiner(self._orm, idquery, fltr).autojoin()
            idquery = idquery.filter(fltr)
        # Only query IDs first
        shapeids = idquery.all()
        def query_factory():
            query = self._session.query(Shape)
            if prefetch_points:
                query = query.options(subqueryload('points'))
            return query
        return self._page_query(query_factory, Shape.feed_id, Shape.shape_id, shapeids, batch_size)

    def fare_attribute(self, fare_id, feed_id="", prefetch_fare_rules=True):
        query = self._session.query(FareAttribute)
        if prefetch_fare_rules:
            query = query.options(subqueryload('fare_rules'))
        return query.get((feed_id, fare_id))

    def fare_attributes(self, fltr=None, prefetch_fare_rules=True):
        query = self._session.query(FareAttribute).distinct()
        if fltr is not None:
            # TODO _AutoJoiner will not work here
            # The join / filter to deduce from a general filter expression
            # is rather complex, as there are many paths from a fare attribute
            # to any object. For example, 4 paths for an agency:
            # Path 1: farerule-route-agency,
            # Path 2/3/4: farerule-zone[origin,destination,contains]-stop-stoptime-trip-route-agency
            # BUT we also sometimes need to include any fare_attributes containing rules that does
            # not correspond to any entities, as they are the default...
            # query = query.join(FareRule).join(Route, FareRule.route).filter(fltr)
            query = query.filter(fltr)
        if prefetch_fare_rules:
            query = query.options(subqueryload('fare_rules'))
        return query.all()

    def fare_rules(self, fltr=None, prefetch_fare_attributes=True):
        query = self._session.query(FareRule).distinct()
        if fltr is not None:
            # TODO _AutoJoiner will not work here - same remark as above.
            # query = query.join(Route, FareRule.route).filter(fltr)
            query = query.filter(fltr)
        if prefetch_fare_attributes:
            query = query.options(subqueryload('fare_attribute'))
        return query.all()

    def _page_query(self, query_factory, item_feed_id_column, item_id_column, ids, batch_size):
        if batch_size <= 0:
            batch_size = 1000
        for feed_id, item_ids in group_pairs(ids, batch_size):
            query = query_factory()
            query = query.filter((item_feed_id_column == feed_id) & (item_id_column.in_(item_ids)))
            batch = query.all()
            for item in batch:
                yield item

    def load_gtfs(self, filename, feed_id="", lenient=False, disable_normalization=False, **kwargs):
        @transactional(self.session())
        def _do_load_gtfs():
            with Gtfs(ZipFileSource(filename)).load() as gtfs:
                _convert_gtfs_model(feed_id, gtfs, self, lenient, disable_normalization, **kwargs)
        _do_load_gtfs()

class _AutoJoiner(object):

    def __init__(self, orm, query, fltr):
        self._orm = orm
        self._query = query
        self._fltr = fltr

    def autojoin(self):
        # 1. Determine the set of classes used in the query
        #    Usually, only one.
        query_classes = set()
        for col_desc in self._query.column_descriptions:
            _type = col_desc.get('type')
            if isclass(_type):
                clazz = _type
            else:
                expr = col_desc.get('expr')
                if hasattr(expr, 'class_'):
                    clazz = expr.class_
                else:
                    print("*** TODO Unrecognized class for expression: %s" % col_desc)
                    clazz = None
            if clazz is not None:
                query_classes.add(clazz)
        # 2. Determine the set of classes used in the filter
        self._join_tables = set()
        self._recurse_inspect(self._fltr)
        join_classes = set()
        for tbl in self._join_tables:
            join_class = self._orm.class_for_table(tbl)
            if join_class is not None:
                join_classes.add(join_class)
            else:
                # TODO Should we handle this?
                # print("Unknown class for table %s" % tbl)
                pass
        for clazz in query_classes:
            if clazz in join_classes:
                join_classes.remove(clazz)
        # 3. Compute all classes (query and join)
        all_classes = set()
        all_classes |= query_classes
        all_classes |= join_classes

        # 4. Ensure the join are connected
        #    Here we use the fact that the relationship graph has only 4 branches:
        #    Branch 1 is agency-route-trip
        #    Branch 2 is dates-calendar-trip
        #    Branch 3 is stops-stoptimes-trip
        #    Branch 4 is shape-trip
        #    With all branches in a star-like configuration with trip in the middle.
        branch1 = Agency in all_classes or Route in all_classes
        branch2 = CalendarDate in all_classes or Calendar in all_classes
        branch3 = Transfer in all_classes or Stop in all_classes or StopTime in all_classes
        branch4 = Shape in all_classes
        n_branches = sum(branch for branch in (branch1, branch2, branch3, branch4))
        if Trip not in all_classes and n_branches > 1:
            join_classes.add(Trip)
            all_classes.add(Trip)
        if Trip in all_classes:
            # Connect branch leafs to trip eventually
            if Agency in all_classes and not Route in all_classes:
                join_classes.add(Route)
            if CalendarDate in all_classes and not Calendar in all_classes:
                join_classes.add(Calendar)
            if (Transfer in all_classes or Stop in all_classes) and not StopTime in all_classes:
                join_classes.add(StopTime)
            if Transfer in all_classes and not Stop in all_classes:
                join_classes.add(Stop)
        else:
            # Connect branch 3, adding Stop if needed
            if Transfer in all_classes and StopTime in all_classes and not Stop in all_classes:
                join_classes.add(Stop)

        # The order of join is important. We could have devised
        # a more generic approach by computing the paths between
        # each join to the query class, but since the number of
        # classes to join is really small we use a brute-force
        # approach: try to join each class in turn and if that
        # fails, re-queue it at the end and try another one...
        # Sort the classes in any order for stability of return values
        classes_to_join = sorted([ clazz for clazz in join_classes ], key=lambda clazz: str(clazz))
        timeout = 0
        while classes_to_join:
            clazz = classes_to_join.pop(0)
            timeout += 1
            try:
                self._query = self._query.join(clazz)
            except InvalidRequestError:
                if timeout > 20:
                    raise
                # Try later on
                classes_to_join.append(clazz)

        return self._query

    def _recurse_inspect(self, fltr_node):
        if hasattr(fltr_node, "table"):
            self._join_tables.add(fltr_node.table.name)
        for child in fltr_node.get_children():
            self._recurse_inspect(child)

def transactional(session):
    def wrap(func):
        def wrapped_func(*args, **kwargs):
            # No need to open the transaction
            # as it is automatically opened in
            # non auto-commit mode.
            try:
                ret = func(*args, **kwargs)
                session.commit()
                return ret
            except:
                session.rollback()
                raise
        return wrapped_func
    return wrap



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""
import math
import pyqtree

# Radius of earth in meters
EARTH_RADIUS = 6371000

def orthodromic_distance(a, b):
    # Use Haversine formula
    lon_a, lat_a, lon_b, lat_b = map(math.radians, [a.lon(), a.lat(), b.lon(), b.lat()])
    dlon = lon_b - lon_a 
    dlat = lat_b - lat_a 
    c = 2 * math.asin(math.sqrt(math.sin(dlat / 2) ** 2 + math.cos(lat_a) * math.cos(lat_b) * math.sin(dlon / 2) ** 2))
    return c * EARTH_RADIUS

"""
@return A 2-tuple composed of
            1) the distance in meter from the point p to the segment [ab],
            2) the distance in meter from a to the clamped projection of p on [ab].
            The clamped projection is the standard projected point if it lies on the
            segment, or one of the extremities of the segment if not.
"""
def orthodromic_seg_distance(p, a, b):
    # Use approximate equirectangular projection
    x_p = math.radians(p.lat())
    cos_p = math.cos(x_p)
    y_p, y_a, y_b = map(lambda q: math.radians(q.lon()) * cos_p, [p, a, b])
    x_a, x_b = map(lambda q: math.radians(q.lat()), [a, b])
    # Compute [AB] length
    l2 = (x_a - x_b) * (x_a - x_b) + (y_a - y_b) * (y_a - y_b)
    if l2 == 0:
        # Pathological d(AB)=0 case, d = d(PA)
        d2 = (x_p - x_a) * (x_p - x_a) + (y_p - y_a) * (y_p - y_a)
        # Any value of t will do
        t = 0
    else:
        # Compute t, linear coordinate of C in the [AB] vector basis
        # and where C is the projection of P on line (AB).
        t = ((x_p - x_a) * (x_b - x_a) + (y_p - y_a) * (y_b - y_a)) / l2
        if t < 0:
            # C outside [AB] on A side: d = d(PA)
            d2 = (x_p - x_a) * (x_p - x_a) + (y_p - y_a) * (y_p - y_a)
            # Clamp
            t = 0
        elif t > 1:
            # C outside [AB] on B side: d = d(PB)
            d2 = (x_p - x_b) * (x_p - x_b) + (y_p - y_b) * (y_p - y_b)
            # Clamp
            t = 1
        else:
            # C inside [AB]: d = d(PC), C = A + t.B
            xC = x_a + t * (x_b - x_a)
            yC = y_a + t * (y_b - y_a)
            d2 = (x_p - xC) * (x_p - xC) + (y_p - yC) * (y_p - yC)
    return EARTH_RADIUS * math.sqrt(d2), EARTH_RADIUS * math.sqrt(l2) * t

class DistanceCache(object):
    
    def __init__(self):
        self._cache = {}
    
    def orthodromic_distance(self, a, b):
        key = (a.lat(), a.lon(), b.lat(), b.lon())
        if key in self._cache:
            return self._cache[key]
        d = orthodromic_distance(a, b)
        self._cache[key] = d
        return d

class RectangularArea(object):
    
    def __init__(self, min_lat, min_lon, max_lat, max_lon):
        self.min_lat = min_lat
        self.min_lon = min_lon
        self.max_lat = max_lat
        self.max_lon = max_lon

    def __repr__(self):
        return "<%s(%f,%f)-(%f,%f)>" % (
                self.__class__.__name__, self.min_lat, self.min_lon, self.max_lat, self.max_lon)
        
# TODO Create circular area (=center+radius)?

class SpatialCluster(object):

    def __init__(self, ident, items):
        self.id = ident
        self.items = items
        self._lat = 0.0
        self._lon = 0.0
        w = 0.0
        for item in self.items:
            self._lat += item.lat()
            self._lon += item.lon()
            w += 1
        self._lat /= w
        self._lon /= w

    def lat(self):
        """The latitude of the cluster barycenter."""
        return self._lat

    def lon(self):
        """The longitude of the cluster barycenter."""
        return self._lon

    def aggregate(self, func, sep=' '):
        """Aggregate distinct values of items in a single string
           (a field or any computed value; such as the item name).
           Example:  cluster.aggregate(lambda s: s.stop_name).
           func should return a string value."""
        vals = list(set([ func(item) for item in self.items ]))
        vals.sort()
        return sep.join(vals)

    def __repr__(self):
        return "<%s(#%d, %s)>" % (
                self.__class__.__name__, self.id, self.items)

class SpatialClusterizer(object):
    """This class is meant to group stops in clusters based on distance proximity.
       It will group all stops that are nearer than D0 meters."""

    def __init__(self, D0):
        self._D0 = 1. * D0
        bbox = (-180, -90, 180, 90)
        self._spidx = pyqtree.Index(bbox)
        self._points = []
        self._clusters_by_item = None
        self._clusters = None

    def add_point(self, p):
        if self._points is None:
            # Well, technically we could, but at the expense of some code complexity.
            # As far as I understand, it's not really needed to clusterize on the fly.
            raise Exception("Can't add point after clusterize() has been called")
        self._spidx.insert(p, (p.lon(), p.lat(),
                               p.lon(), p.lat()))
        self._points.append(p)

    def add_points(self, ps):
        for p in ps:
            self.add_point(p)

    @staticmethod
    def def_comparator(d, d0, p1, p2):
        """Consider 2 points to be in the same cluster if their distance is closer
           than the original threshold D0. Do not take into account anything else."""
        return True

    def make_comparator(self, same_name=False, different_station_penalty=0.5):
        """Build a comparator that apply the following rules:
           - If samename, force the stops to have the same name to be clusterized
           - Apply the different_station_penalty coefficient to the distance when
             the stops do not belong to the same station. Must be <1.0"""
        def stop_station_comparator(d, d0, stop1, stop2):
            if same_name and stop1.stop_name != stop2.stop_name:
                return False
            penalty = 1.0
            if different_station_penalty < penalty and not stop1.in_same_station(stop2):
                penalty = different_station_penalty
            return d <= d0 * penalty
        return stop_station_comparator

    def clusterize(self, comparator=def_comparator.__func__):
        clusters_by_item = {}
        for p in self._points:
            clusters_by_item[p] = set([p])
        for p1 in self._points:
            d0 = self._D0
            # Compute bounds in lat,lon degree of d square around stop
            cos_lat = math.cos(math.radians(p.lat()))
            dlon = math.degrees(d0 / EARTH_RADIUS * cos_lat)
            dlat = math.degrees(d0 / EARTH_RADIUS)
            bbox = (p1.lon() - dlon, p1.lat() - dlat,
                    p1.lon() + dlon, p1.lat() + dlat)
            nearby = self._spidx.intersect(bbox)
            for p2 in nearby:
                d = orthodromic_distance(p1, p2)
                if d > d0:
                    continue
                if not comparator(d, self._D0, p1, p2):
                    continue
                c1 = clusters_by_item.get(p1)
                c2 = clusters_by_item.get(p2)
                if c1 is c2:
                    # Already same cluster
                    continue
                # Merge cluster c1 and c2
                c3 = c1 | c2
                for p in c3:
                    clusters_by_item[p] = c3
        seen_points = set()
        self._clusters_by_item = {}
        self._clusters = []
        cluster_id = 0
        for _k, cluster in clusters_by_item.items():
            to_process = True
            for p in cluster:
                if p in seen_points:
                    to_process = False
                    break
                seen_points.add(p)
            if not to_process:
                continue
            spatial_cluster = SpatialCluster(cluster_id, cluster)
            cluster_id += 1
            for p in cluster:
                self._clusters_by_item[p] = spatial_cluster
            self._clusters.append(spatial_cluster)
        # None needed anymore
        self._points = None

    def cluster_of(self, p):
        return self._clusters_by_item.get(p)

    def in_same_cluster(self, p1, p2):
        c1 = self._clusters_by_item.get(p1)
        c2 = self._clusters_by_item.get(p2)
        if c1 is None or c2 is None:
            return False
        return c1 is c2

    def clusters(self):
        return self._clusters



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from _collections import defaultdict
from gtfslib.utils import fmttime

"Cf. plugin pour la documentation"
def decret_2015_1610(trips, trace=True, required_distance=500, required_ratio=2.5):

    affiche(trace, "Calcul decret 2015 1610 sur %d voyages." % (len(trips)))
    if len(trips) == 0:
        affiche(trace, "Aucun voyages, impossible de calculer.")
        return None, None, None

    affiche(trace, "Calcul de l'espacement moyen des arrêts...")
    espacement_moyen = 0
    w_esp = 0
    for trip in trips:
        # Note: on pondère par le nombre de jours chaque voyage est applicable.
        # Ceci permet de prendre en compte la fréquence: par exemple, la distance
        # d'un intervalle entre deux arrêt actif le lundi uniquement sera pondéré
        # 5 fois moins qu'un autre intervalle actif du lundi au vendredi.
        n_jours = len(trip.calendar.dates)
        for stoptime1, stoptime2 in trip.hops():
            espacement_moyen += (stoptime2.shape_dist_traveled - stoptime1.shape_dist_traveled) * n_jours
            w_esp += n_jours
    espacement_moyen /= w_esp
    affiche(trace, "L'espacement moyen entre arrêt du réseau est de %.2f mètres (max %.0fm)." % (espacement_moyen, float(required_distance)))

    affiche(trace, "Calcul du jour ayant la fréquence en voyage la plus élevée...")
    frequences = defaultdict(lambda: 0)
    for trip in trips:
        for date in trip.calendar.dates:
            frequences[date] += 1
    date_max = None
    freq_max = 0
    for date, frequence in frequences.items():
        if frequence > freq_max:
            freq_max = frequence
            date_max = date
    affiche(trace, "Le jour ayant le nombre de voyage le plus élevé est le %s, avec %d voyages." % (date_max.as_date(), freq_max))

    affiche(trace, "Calcul des fréquences sur la plage horaire 8h - 19h...")
    # TODO Est-ce que ce calcul est correct? Le décret est pas clair.
    # On calcule le nombre de voyages actifs pendant chaque minute.
    frequences = [ 0 for minute in range(0, 20 * 60) ]
    for trip in trips:
        if date_max not in trip.calendar.dates:
            continue
        minute_depart = trip.stop_times[0].departure_time // 60
        minute_arrivee = trip.stop_times[-1].arrival_time // 60
        for minute in range(minute_depart, minute_arrivee + 1):
            if minute >= 8 * 60 and minute < 20 * 60:
                frequences[minute] += 1
    frequence_min = 99999999999
    minute_min = 0
    frequence_max = 0
    minute_max = 0
    # La fréquence horaire min/max est calculé en moyenne glissante
    # sur une heure en sommant les fréquences par minute.
    for minute in range(8 * 60, 19 * 60):
        freq = 0
        for delta_minute in range(0, 60):
            freq += frequences[minute + delta_minute]
        if freq > frequence_max:
            frequence_max = freq
            minute_max = minute
        if freq < frequence_min:
            frequence_min = freq
            minute_min = minute
    affiche(trace, "La fréquence minimale est de %.2f voyages/heure, entre %s et %s." % (frequence_min / 60.0, fmttime(minute_min * 60), fmttime((minute_min + 60) * 60)))
    affiche(trace, "La fréquence maximale est de %.2f voyages/heure, entre %s et %s." % (frequence_max / 60.0, fmttime(minute_max * 60), fmttime((minute_max + 60) * 60)))
    if frequence_min == 0:
        ratio_frequence = float('inf')
    else:
        ratio_frequence = frequence_max / float(frequence_min)
    affiche(trace, "Le ratio entre fréquence max et min est de %.3f (max %.2f)." % (ratio_frequence, float(required_ratio)))

    urbain = ratio_frequence < required_ratio and espacement_moyen < required_distance
    affiche(trace, "Ce service est %s au sens du décret n° 2015-1610."
          % ("URBAIN" if urbain else "NON URBAIN"))
    return urbain, espacement_moyen, ratio_frequence

def affiche(affiche, message):
    if affiche:
        print(message)

class Decret_2015_1610(object):
    """
    Teste un ensemble de voyages (trips) par rapport au décret n° 2015-1610
    du 8 décembre 2015 relatif aux critères d'espacement moyen des arrêts
    et de variation de la fréquence de passage des services réguliers de
    transport public routier urbain de personnes. Pour plus d'informations:
    http://www.legifrance.gouv.fr/affichTexte.do?cidTexte=JORFTEXT000031589954

    Paramètres:
    --distance=<dist>   Distance minimale moyenne entre arrêts.
                        Valeur par défaut (décret): 500m
    --ratio=<ratio>     Ratio entre fréquence horaire minimale et maximale.
                        Valeur par défaut (décret): 2.5
    """
    def run(self, context, distance=500, ratio=2.5, **kwargs):
        print("Chargement des données...")
        trips = context.dao().trips(fltr=context.args.filter, prefetch_stop_times=True, prefetch_calendars=True, prefetch_stops=False)
        trips = list(trips)
        urbain = decret_2015_1610(trips, trace=True, required_distance=float(distance), required_ratio=float(ratio))
        return urbain



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""
import bisect
import logging
import time
import six.moves
from collections import defaultdict

def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        logging.info("%s() took %0.3f sec" % (f.__name__, time2 - time1))
        return ret
    return wrap

def gtfstime(h, m, s=0):
    return h * 3600 + m * 60 + s

def fmttime(ssm):
    if ssm is None:
        return ""
    h = ssm / 3600
    ssm %= 3600
    m = ssm / 60
    ssm %= 60
    s = ssm
    return "%d:%02d:%02d" % (h, m, s)

def group_items(items, group_size):
    """Split items into chunks of size group_size
       (except the last which can obviously be shorter).
       Unbelievably, this is not part of standard python."""
    itemlist = list(items)
    for idx in six.moves.range(0, len(itemlist), group_size):
        yield itemlist[idx:idx + group_size]

def group_pairs(pairs, group_size):
    groups = defaultdict(list)
    for a, b in pairs:
        groups[a].append(b)
    for a, bs in groups.items():
        for idx in six.moves.range(0, len(bs), group_size):
            yield a, bs[idx:idx + group_size]

class ContinousPiecewiseLinearFunc(object):

    def __init__(self):
        self._x = []
        self._y = []
        self._sorted = True

    def append(self, x, y):
        self._x.append(x)
        self._y.append(y)
        self._sorted = False

    def interpolate(self, x):

        if len(self._x) == 0:
            raise Exception("Empty piecewise linear function")

        if not self._sorted:
            zipped = zip(self._x, self._y)
            zipped = sorted(zipped)
            self._x = [ ax for (ax, ay) in zipped ]
            self._y = [ ay for (ax, ay) in zipped ]
            self._sorted = True

        idx = bisect.bisect(self._x, x)
        if idx == 0:
            # Clamp to left
            return self._y[0]
        if idx == len(self._x):
            # Clamp to right
            return self._y[-1]
        x1 = self._x[idx - 1]
        x2 = self._x[idx]
        y1 = self._y[idx - 1]
        y2 = self._y[idx]
        dx = x2 - x1
        # Please note that by construction dx can never be null
        # bisect always return the last item in case several are equals
        dy = y2 - y1
        return (1.0 * x - x1) * dy / dx + y1



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from gtfsplugins.prettycsv import PrettyCsv
from collections import defaultdict

class TripsPerDay(object):
    """
    Compute the number of trips per route per day.

    Parameters:
    --csv=<file>    Output to given file as CSV
    --maxwidth=<n>  Max column width of console output
    --byagency      Aggregate by agency
    --byroute       Aggregate by route
    --bydir         Aggregate by route and direction

    Examples:
    --filter="(CalendarDate.date >= '2016-01-01') &
              (CalendarDate.date <= '2016-01-31')"
      Restrict to some date range
    --filter="(Route.route_short_name=='R1')"
      Restrict to route R1
    """

    def __init__(self):
        pass

    def run(self, context, csv=None, byagency=False, byroute=False, bydir=False, **kwargs):

        print("Loading calendar dates...")
        dates = set(context.dao().calendar_dates_date(fltr=context.args.filter))
        print("Loaded %d dates" % (len(dates)))

        print("Processing trips...")
        tripcount = defaultdict(lambda: defaultdict(int))
        keys = set()
        ntrips = 0
        for trip in context.dao().trips(fltr=context.args.filter, prefetch_stops=False, prefetch_stop_times=False, prefetch_calendars=True, prefetch_routes=True):
            if byagency:
                key = (trip.route.agency.agency_name, trip.route.agency)
            elif byroute:
                key = (trip.route.name(), trip.route)
            elif bydir:
                key = ("%s-%d" % (trip.route.name(), trip.direction_id), (trip.direction_id, trip.route))
            else:
                key = ("SUM", None)
            keys.add(key)
            for date in trip.calendar.dates:
                if date.as_date() not in dates:
                    continue
                tripcount[date.as_date()][key] += 1
            if ntrips % 1000 == 0:
                print("%d trips..." % (ntrips))
            ntrips += 1
        print("Processed %d trips" % (ntrips))

        keys = list(keys)
        keys.sort(key=lambda k: k[0])
        dates = list(tripcount.keys())
        dates.sort()

        with PrettyCsv(csv, ["date"] + [ key[0] for key in keys ], **kwargs) as csvout:
            for date in dates:
                row = { "date" : date }
                tkc = tripcount.get(date)
                for key in keys:
                    count = tkc.get(key, 0)
                    if count > 0:
                        row[key[0]] = count
                csvout.writerow(row)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

from gtfslib.utils import ContinousPiecewiseLinearFunc, group_items, group_pairs
import unittest

class TestUtils(unittest.TestCase):

    def test_piecewise_linear_interpolation(self):

        # Empty function
        f = ContinousPiecewiseLinearFunc()
        catched = False
        try:
            self.assertAlmostEqual(f.interpolate(0), 0, 3)
        except:
            catched = True
        self.assertTrue(catched)

        # Single point function
        f = ContinousPiecewiseLinearFunc()
        f.append(0, 0)
        self.assertAlmostEqual(f.interpolate(0), 0, 6)
        self.assertAlmostEqual(f.interpolate(-1), 0, 6)
        self.assertAlmostEqual(f.interpolate(1), 0, 6)

        # A simple well-behaved function
        f = ContinousPiecewiseLinearFunc()
        f.append(0, 0)
        f.append(1, 100)
        self.assertAlmostEqual(f.interpolate(-0.000000001), 0, 6)
        self.assertAlmostEqual(f.interpolate(0), 0, 6)
        self.assertAlmostEqual(f.interpolate(0.000000001), 0, 6)
        self.assertAlmostEqual(f.interpolate(0.5), 50.0, 6)
        self.assertAlmostEqual(f.interpolate(0.9999999999), 100, 6)
        self.assertAlmostEqual(f.interpolate(1.0), 100, 6)
        self.assertAlmostEqual(f.interpolate(1.0000000001), 100, 6)

        # Function with double-point (infinite derivative)
        f = ContinousPiecewiseLinearFunc()
        f.append(0, 0)
        f.append(1, 0)
        f.append(1, 100)
        f.append(2, 100)
        self.assertAlmostEqual(f.interpolate(0), 0, 6)
        self.assertAlmostEqual(f.interpolate(1), 100, 6)
        self.assertAlmostEqual(f.interpolate(-1), 0, 6)

        # A more complex one
        f = ContinousPiecewiseLinearFunc()
        f.append(10, 9000)
        f.append(20, 10000)
        f.append(1000, 10980)
        self.assertAlmostEqual(f.interpolate(5), 9000.0, 6)
        self.assertAlmostEqual(f.interpolate(15), 9500.0, 6)
        self.assertAlmostEqual(f.interpolate(40), 10020.0, 3)
        self.assertAlmostEqual(f.interpolate(1010), 10980.0, 3)

    def test_group_items(self):
        for alist in [ range(0), range(5), range(10), range(15), range(100) ]:
            alist2 = []
            for aa in group_items(alist, 10):
                for a in aa:
                    alist2.append(a)
            self.assertTrue(list(alist) == alist2)

    def test_group_pairs(self):

        pairs = list(group_pairs([ ], 100))
        self.assertTrue(len(pairs) == 0)

        pairs = list(group_pairs([ ('A', 1), ('A', 2) ], 100))
        self.assertTrue(len(pairs) == 1)
        self.assertTrue(pairs == [('A', [1, 2])])

        pairs = sorted(list(group_pairs([ ('A', 1), ('B', 2) ], 100)))
        self.assertTrue(len(pairs) == 2)
        self.assertTrue(pairs == [ ('A', [1]), ('B', [2]) ])

        pairs = sorted(list(group_pairs([ ('A', 1), ('B', 2), ('A', 3), ('A', 4) ], 2)))
        self.assertTrue(len(pairs) == 3)
        self.assertTrue(pairs == [ ('A', [1, 3]), ('A', [4]), ('B', [2]) ])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import datetime
import unittest

from sqlalchemy.sql.expression import or_
from sqlalchemy.sql.functions import func
from sqlalchemy.orm import clear_mappers

from gtfslib.dao import Dao
from gtfslib.model import CalendarDate, Route, Calendar, Stop, \
    Trip, StopTime
from gtfslib.spatial import RectangularArea, SpatialClusterizer
from gtfslib.utils import gtfstime


# Location of dummy.gtfs.zip.
# This unit-test is highly dependent on the CONTENT of this GTFS.
DUMMY_GTFS = "test/dummy.gtfs.zip"

DAO_URL = ""
# To unit-test with postgresql, create a db "gtfs" with user "gtfs" and uncomment the following line:
# DAO_URL = "postgresql://gtfs@localhost/gtfs"
# Set this to true to activate SQL logging
SQL_LOG = False

class TestDummyGtfs(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    def test_gtfs_data(self):
        dao = Dao(DAO_URL, sql_logging=False)
        dao.load_gtfs(DUMMY_GTFS)

        # Check feed
        feed = dao.feed()
        self.assertTrue(feed.feed_id == "")
        self.assertTrue(feed.feed_publisher_name == "Mecatran")
        self.assertTrue(feed.feed_publisher_url == "http://www.mecatran.com/")
        self.assertTrue(feed.feed_contact_email == "support@mecatran.com")
        self.assertTrue(feed.feed_lang == "fr")
        self.assertTrue(len(dao.agencies()) == 2)
        self.assertTrue(len(dao.routes()) == 3)
        self.assertTrue(len(feed.agencies) == 2)
        self.assertTrue(len(feed.routes) == 3)

        # Check agencies
        at = dao.agency("AT")
        self.assertTrue(at.agency_name == "Agency Train")
        self.assertTrue(len(at.routes) == 1)
        ab = dao.agency("AB")
        self.assertTrue(ab.agency_name == "Agency Bus")
        self.assertTrue(len(ab.routes) == 2)

        # Check calendars
        week = dao.calendar("WEEK")
        self.assertTrue(len(week.dates) == 253)
        summer = dao.calendar("SUMMER")
        self.assertTrue(len(summer.dates) == 42)
        mon = dao.calendar("MONDAY")
        self.assertTrue(len(mon.dates) == 49)
        sat = dao.calendar("SAT")
        self.assertTrue(len(sat.dates) == 53)
        for date in mon.dates:
            self.assertTrue(date.dow() == 0)
        for date in sat.dates:
            self.assertTrue(date.dow() == 5)
        for date in week.dates:
            self.assertTrue(date.dow() >= 0 and date.dow() <= 4)
        for date in summer.dates:
            self.assertTrue(date >= CalendarDate.ymd(2016, 7, 1) and date <= CalendarDate.ymd(2016, 8, 31))
        empty = dao.calendars(func.date(CalendarDate.date) == datetime.date(2016, 5, 1))
        # OR USE: empty = dao.calendars(CalendarDate.date == "2016-05-01")
        self.assertTrue(len(empty) == 0)
        july4 = CalendarDate.ymd(2016, 7, 4)
        summer_mon = dao.calendars(func.date(CalendarDate.date) == july4.date)
        n = 0
        for cal in summer_mon:
            self.assertTrue(july4 in cal.dates)
            n += 1
        self.assertTrue(n == 3)

        # Check stops
        sbq = dao.stop("BQ")
        self.assertAlmostEqual(sbq.stop_lat, 44.844, places=2)
        self.assertAlmostEqual(sbq.stop_lon, -0.573, places=2)
        self.assertTrue(sbq.stop_name == "Bordeaux Quinconces")
        n = 0
        for stop in dao.stops(Stop.stop_name.like("Gare%")):
            self.assertTrue(stop.stop_name.startswith("Gare"))
            n += 1
        self.assertTrue(n == 7)
        n = 0
        for stop in dao.stops(fltr=dao.in_area(RectangularArea(44.7, -0.6, 44.9, -0.4))):
            self.assertTrue(stop.stop_lat >= 44.7 and stop.stop_lat <= 44.9 and stop.stop_lon >= -0.6 and stop.stop_lon <= -0.4)
            n += 1
        self.assertTrue(n == 16)
        for station in dao.stops(Stop.location_type == Stop.TYPE_STATION):
            self.assertTrue(station.location_type == Stop.TYPE_STATION)
            self.assertTrue(len(station.sub_stops) >= 2)
            for stop in station.sub_stops:
                self.assertTrue(stop.parent_station == station)

        # Check zones
        z_inexistant = dao.zone("ZX")
        self.assertTrue(z_inexistant is None)
        z1 = dao.zone("Z1")
        self.assertEquals(16, len(z1.stops))
        z2 = dao.zone("Z2")
        self.assertEquals(4, len(z2.stops))

        # Check transfers
        transfers = dao.transfers()
        self.assertTrue(len(transfers) == 3)
        transfers = dao.transfers(fltr=(dao.transfer_from_stop().stop_id == 'GBSJB'))
        self.assertTrue(len(transfers) == 1)
        self.assertTrue(transfers[0].from_stop.stop_id == 'GBSJB')

        # Check routes
        tgv = dao.route("TGVBP")
        self.assertTrue(tgv.agency == at)
        self.assertTrue(tgv.route_type == 2)
        r1 = dao.route("BR")
        self.assertTrue(r1.route_short_name == "R1")
        self.assertTrue(r1.route_long_name == "Bus Red")
        n = 0
        for route in dao.routes(Route.route_type == 3):
            self.assertTrue(route.route_type == 3)
            n += 1
        self.assertTrue(n == 2)

        # Check trip for route
        n = 0
        trips = dao.trips(fltr=Route.route_type == Route.TYPE_BUS)
        for trip in trips:
            self.assertTrue(trip.route.route_type == Route.TYPE_BUS)
            n += 1
        self.assertTrue(n > 20)

        # Check trips on date
        trips = dao.trips(fltr=func.date(CalendarDate.date) == july4.date, prefetch_calendars=True)
        n = 0
        for trip in trips:
            self.assertTrue(july4 in trip.calendar.dates)
            n += 1
        self.assertTrue(n > 30)

    def test_non_overlapping_feeds(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        # Load twice the same data under two distinct namespaces
        dao.load_gtfs(DUMMY_GTFS, feed_id='A')
        dao.load_gtfs(DUMMY_GTFS, feed_id='B')

        # Check that each feed only return it's own data
        feed_a = dao.feed('A')
        self.assertTrue(feed_a.feed_id == 'A')
        feed_b = dao.feed('B')
        self.assertTrue(feed_b.feed_id == 'B')
        self.assertTrue(len(dao.agencies()) == 4)
        self.assertTrue(len(feed_a.agencies) == 2)
        self.assertTrue(len(feed_b.agencies) == 2)
        self.assertTrue(len(feed_a.routes) * 2 == len(dao.routes()))
        self.assertTrue(len(feed_b.routes) * 2 == len(dao.routes()))
        self.assertTrue(len(feed_a.stops) * 2 == len(list(dao.stops())))
        self.assertTrue(len(feed_b.stops) * 2 == len(list(dao.stops())))
        self.assertTrue(len(feed_a.calendars) * 2 == len(dao.calendars()))
        self.assertTrue(len(feed_b.calendars) * 2 == len(dao.calendars()))
        self.assertTrue(len(feed_a.trips) * 2 == len(list(dao.trips())))
        self.assertTrue(len(feed_b.trips) * 2 == len(list(dao.trips())))

    def test_complex_queries(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(DUMMY_GTFS)

        # Get the list of departures:
        # 1) from "Porte de Bourgogne"
        # 2) on 4th July
        # 3) between 10:00 and 14:00
        # 4) on route type BUS
        # 5) not the last of trip (only departing)
        porte_bourgogne = dao.stop("BBG")
        july4 = CalendarDate.ymd(2016, 7, 4)
        from_time = gtfstime(10, 00)
        to_time = gtfstime(14, 00)
        departures = dao.stoptimes(
                    fltr=(StopTime.stop == porte_bourgogne) & (StopTime.departure_time >= from_time) & (StopTime.departure_time <= to_time)
                    & (Route.route_type == Route.TYPE_BUS) & (func.date(CalendarDate.date) == july4.date),
                    prefetch_trips=True)

        n = 0
        for dep in departures:
            self.assertTrue(dep.stop == porte_bourgogne)
            self.assertTrue(july4 in dep.trip.calendar.dates)
            self.assertTrue(dep.trip.route.route_type == Route.TYPE_BUS)
            self.assertTrue(dep.departure_time >= from_time and dep.departure_time <= to_time)
            n += 1
        self.assertTrue(n > 10)

        # Plage is a stop that is used only in summer (hence the name!)
        plage = dao.stop("BPG")
        # Get the list of stops used by some route:
        # 1) All-year round
        route_red = dao.route("BR")
        stoplist_all = list(dao.stops(fltr=Trip.route == route_red))
        # 2) Only in january
        from_date = CalendarDate.ymd(2016, 1, 1)
        to_date = CalendarDate.ymd(2016, 1, 31)
        stoplist_jan = list(dao.stops(
                    fltr=(Trip.route == route_red) & (func.date(CalendarDate.date) >= from_date.date) & (func.date(CalendarDate.date) <= to_date.date)))
        # Now, make some tests
        self.assertTrue(len(stoplist_all) > 5)
        self.assertTrue(plage in stoplist_all)
        self.assertFalse(plage in stoplist_jan)
        stoplist = list(stoplist_all)
        stoplist.remove(plage)
        self.assertTrue(set(stoplist) == set(stoplist_jan))

        # Get all routes passing by the set of stops
        routes = dao.routes(fltr=or_(StopTime.stop == stop for stop in stoplist_jan))
        stopset = set()
        for route in routes:
            for trip in route.trips:
                for stoptime in trip.stop_times:
                    stopset.add(stoptime.stop)
        self.assertTrue(set(stoplist_jan).issubset(stopset))

    def test_custom_queries(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(DUMMY_GTFS)

        # A simple custom query: count the number of stops per type (stop/station)
        # SQL equivalent: SELECT stop.location_type, count(stop.location_type) FROM stop GROUP BY stop.location_type
        for type, stop_count in dao.session() \
                    .query(Stop.location_type, func.count(Stop.location_type)) \
                    .group_by(Stop.location_type) \
                   .all():
            # print("type %d : %d stops" % (type, stop_count))
            if type == Stop.TYPE_STATION:
                self.assertTrue(stop_count == 3)
            if type == Stop.TYPE_STOP:
                self.assertTrue(stop_count > 15 and stop_count < 30)

        # A more complex custom query: count the number of trips per calendar date per route on june/july
        from_date = CalendarDate.ymd(2016, 6, 1)
        to_date = CalendarDate.ymd(2016, 7, 31)
        for date, route, trip_count in dao.session() \
                    .query(CalendarDate.date, Route, func.count(Trip.trip_id)) \
                    .join(Calendar).join(Trip).join(Route) \
                    .filter((func.date(CalendarDate.date) >= from_date.date) & (func.date(CalendarDate.date) <= to_date.date)) \
                    .group_by(CalendarDate.date, Route.route_short_name) \
                    .all():
            # print("%s / %20s : %d trips" % (date, route.route_short_name + " " + route.route_long_name, trip_count))
            self.assertTrue(date >= from_date.as_date())
            self.assertTrue(date <= to_date.as_date())
            self.assertTrue(trip_count > 0)

    def test_clusterizer(self):
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(DUMMY_GTFS)

        # Merge stops closer than 300m together
        sc = SpatialClusterizer(300.0)
        for stop in dao.stops():
            sc.add_point(stop)
        sc.clusterize()

        # for cluster in sc.clusters():
        #    print("---CLUSTER: %d stops" % (len(cluster)))
        #    for stop in cluster:
        #        print("%s %s" % (stop.stop_id, stop.stop_name))

        gare1 = dao.stop("GBSJT")
        gare2 = dao.stop("GBSJ")
        gare3 = dao.stop("GBSJB")
        self.assertTrue(sc.in_same_cluster(gare1, gare2))
        self.assertTrue(sc.in_same_cluster(gare1, gare3))
        self.assertTrue(sc.in_same_cluster(gare2, gare3))

        bq = dao.stop("BQ")
        bq1 = dao.stop("BQA")
        bq2 = dao.stop("BQD")
        self.assertTrue(sc.in_same_cluster(bq, bq1))
        self.assertTrue(sc.in_same_cluster(bq, bq2))

        bs = dao.stop("BS")
        bs1 = dao.stop("BS1")
        bs2 = dao.stop("BS2")
        self.assertTrue(sc.in_same_cluster(bs, bs1))
        self.assertTrue(sc.in_same_cluster(bs, bs2))

        self.assertFalse(sc.in_same_cluster(gare1, bq))
        self.assertFalse(sc.in_same_cluster(gare1, bs))
        self.assertFalse(sc.in_same_cluster(gare3, bs2))

        bjb = dao.stop("BJB")
        self.assertFalse(sc.in_same_cluster(bjb, gare1))
        self.assertFalse(sc.in_same_cluster(bjb, bs))
        self.assertFalse(sc.in_same_cluster(bjb, bq))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""
import hashlib
import traceback
import logging
import os
import random
import unittest
import requests

from sqlalchemy.orm import clear_mappers

from gtfslib.dao import Dao

# By default we do not enable this test, it takes ages
ENABLE = False
# Skip import of already downloaded GTFS
# Handy for re-launching the test w/o having to redo all
SKIP_EXISTING = True
# Limit loading to small GTFS only. Max size in bytes
MAX_GTFS_SIZE = 2 * 1024 * 1024
# Local cache directory
DIR = "all-gtfs.cache"
# List of ID to load. If none, download the whole list
# and process it at random.
IDS_TO_LOAD = None

# The following were known to have fancy formats that used to break:
# Contains UTF-8 BOM:
# IDS_TO_LOAD = [ 'janesville-transit-system' ]
# Contain header with space
# IDS_TO_LOAD = [ 'rseau-stan' ]
# Breaks on non-unique stop time
# IDS_TO_LOAD = [ 'biaostocka-komunikacja-miejska' ]

class TestAllGtfs(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    # Downlaod all GTFS from GTFS data-exchange web-site
    # and load them into a DAO.
    def test_all_gtfs(self):

        if not ENABLE:
            print("This test is disabled as it is very time-consuming.")
            print("If you want to enable it, please see in the code.")
            return

        # Create temporary directory if not there
        if not os.path.isdir(DIR):
            os.mkdir(DIR)

        # Create a DAO. Re-use any existing present.
        logging.basicConfig(level=logging.INFO)
        dao = Dao("%s/all_gtfs.sqlite" % (DIR))

        deids = IDS_TO_LOAD
        if deids is None:
            print("Downloading meta-info for all agencies...")
            resource_url = "http://www.gtfs-data-exchange.com/api/agencies?format=json"
            response = requests.get(resource_url).json()
            if response.get('status_code') != 200:
                raise IOError()
            deids = []
            for entry in response.get('data'):
                deid = entry.get('dataexchange_id')
                deids.append(deid)
            # Randomize the list, otherwise we will always load ABCBus, then ...
            random.shuffle(deids)

        for deid in deids:
            try:
                local_filename = "%s/%s.gtfs.zip" % (DIR, deid)
                if os.path.exists(local_filename) and SKIP_EXISTING:
                    print("Skipping [%s], GTFS already present." % (deid))
                    continue

                print("Downloading meta-info for ID [%s]" % (deid))
                resource_url = "http://www.gtfs-data-exchange.com/api/agency?agency=%s&format=json" % deid
                response = requests.get(resource_url).json()
                status_code = response.get('status_code')
                if status_code != 200:
                    raise IOError("Error %d (%s)" % (status_code, response.get('status_txt')))
                data = response.get('data')
                agency_data = data.get('agency')
                agency_name = agency_data.get('name')
                agency_area = agency_data.get('area')
                agency_country = agency_data.get('country')

                print("Processing [%s] %s (%s / %s)" % (deid, agency_name, agency_country, agency_area))
                date_max = 0.0
                file_url = None
                file_size = 0
                file_md5 = None
                for datafile in data.get('datafiles'):
                    date_added = datafile.get('date_added')
                    if date_added > date_max:
                        date_max = date_added
                        file_url = datafile.get('file_url')
                        file_size = datafile.get('size')
                        file_md5 = datafile.get('md5sum')
                if file_url is None:
                    print("No datafile available, skipping.")
                    continue

                if file_size > MAX_GTFS_SIZE:
                    print("GTFS too large (%d bytes > max %d), skipping." % (file_size, MAX_GTFS_SIZE))
                    continue

                # Check if the file is present and do not download it.
                try:
                    existing_md5 = hashlib.md5(open(local_filename, 'rb').read()).hexdigest()
                except:
                    existing_md5 = None
                if existing_md5 == file_md5:
                    print("Using existing file '%s': MD5 checksum matches." % (local_filename))
                else:
                    print("Downloading file '%s' to '%s' (%d bytes)" % (file_url, local_filename, file_size))
                    with open(local_filename, 'wb') as local_file:
                        cnx = requests.get(file_url, stream=True)
                        for block in cnx.iter_content(1024):
                            local_file.write(block)
                    cnx.close()

                feed = dao.feed(deid)
                if feed is not None:
                    print("Removing existing data for feed [%s]" % (deid))
                    dao.delete_feed(deid)
                print("Importing into DAO as ID [%s]" % (deid))
                try:
                    dao.load_gtfs("%s/%s.gtfs.zip" % (DIR, deid), feed_id=deid)
                except:
                    error_filename = "%s/%s.error" % (DIR, deid)
                    print("Import of [%s]: FAILED. Logging error to '%s'" % (deid, error_filename))
                    with open(error_filename, 'wb') as errfile:
                        errfile.write(traceback.format_exc())
                    raise
                print("Import of [%s]: OK." % (deid))

            except Exception as error:
                logging.exception(error)
                continue

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

class DemoPlugin(object):
    """
    Documentation for the demo plugin.
    This plugin is a demo to illustrate how to implement a plugin.

    Arguments:
      --printtrips Print all filtered trips.
    """

    def __init__(self):
        pass

    def run(self, context, printtrips=False, **kwargs):
        print("%s is running." % (self.__class__.__name__))
        print("Extra arguments: %s" % (kwargs))
        if printtrips:
            for trip in context.dao().trips(fltr=context.args.filter):
                print(trip)


# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import six
import logging
import argparse

from gtfsplugins.demoplugin import DemoPlugin
from gtfsplugins.decret_2015_1610 import Decret_2015_1610
from gtfsplugins.frequencies import Frequencies
from gtfsplugins.tripsperday import TripsPerDay
from gtfsplugins.shpexport import ShapefileExport
from gtfsplugins.export import GtfsExport

# Please keep the following unused imports, they are used by the --filter eval.
from gtfslib.dao import Dao
from gtfslib.model import FeedInfo, Agency, Route, Zone, Stop, Calendar, CalendarDate, Trip, StopTime, Shape, ShapePoint, FareAttribute, FareRule  # @UnusedImport
from gtfslib.spatial import RectangularArea  # @UnusedImport

# TODO Dynamically scan packages
PLUGINS = [ DemoPlugin, Decret_2015_1610, Frequencies, TripsPerDay, ShapefileExport, GtfsExport ]

class PluginContext(object):
    """The class given as execution context to a plugin.
    Propose helper methods to filter objects, or get the DAO."""

    def __init__(self, dao, args):
        self._dao = dao
        self.args = args

    def dao(self):
        return self._dao

def main():

    parser = argparse.ArgumentParser(description='GTFS plugin runner', epilog="Copyright (c) 2016 AFIMB & Mecatran")
    parser.add_argument("database", nargs='?', help="Database to load data from")
    parser.add_argument("plugin", nargs='?', help="Name of plugin to run")
    parser.add_argument('-l', '--list', help="List all available plugins", action="store_true")
    parser.add_argument('--filter', help="Specify object filter. Examples: --filter=\"Stop.stop_name=='FooBar'\",  --filter=\"(Route.route_type.in_([Route.TYPE_SUBWAY, Route.TYPE_TRAM])\",  --filter=\"CalendarDate.date=='2016-02-22'\"", type=str)
    parser.add_argument('--logsql', help="Enable SQL logging (very verbose)", action="store_true")
    parser.add_argument('--schema', help="Set the schema to use (for PostgreSQL).", type=str)

    args, xargs = parser.parse_known_args()

    # Extract extra arguments, to give them to the plugin
    kwargs = {}
    for xarg in xargs:
        kv = xarg.split('=', 1)
        if len(kv) != 1 and len(kv) != 2 or not kv[0].startswith('--'):
            print("Invalid extra argument: %s, use '--param=value' (or '--param' for boolean flag)" % (xarg))
            continue
        kwargs[kv[0][2:]] = kv[1] if len(kv) == 2 else True

    # List plugins
    if args.list:
        print("List of all available plugins:")
        print("")
        for plugin in PLUGINS:
            if plugin.__doc__:
                print("%s:" % plugin.__name__)
                print(plugin.__doc__)
        exit(0)

    # Lookup for plugin
    try:
        plugin_class = six.next(p for p in PLUGINS if p.__name__ == args.plugin)
    except:
        print("Plugin not found: %s" % args.plugin)
        exit(1)

    dao = Dao(args.database, sql_logging=args.logsql, schema=args.schema)

    def _evaluate(strarg, dao):
        return None if strarg is None else eval(strarg, globals(), locals())
    args.filter = _evaluate(args.filter, dao)

    # TODO Configure logging
    logging.basicConfig(level=logging.INFO)

    context = PluginContext(dao, args)

    # Instantiate and run the plugin
    plugin = plugin_class()
    retval = plugin.run(context, **kwargs)

    # Convert the return value to something digestable by scripts
    if isinstance(retval, int):
        exit(retval)
    else:
        exit(1 if retval else 0)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import re
import unittest

from sqlalchemy.orm import clear_mappers

from gtfslib.dao import _AutoJoiner, Dao
from gtfslib.model import CalendarDate, FeedInfo, Agency, Route, Calendar, Stop, \
    Trip, StopTime, Transfer, Shape, ShapePoint, Zone, FareAttribute, FareRule


class TestAutoJoin(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        clear_mappers()

    def test_autojoin(self):
        dao = Dao()

        query = _AutoJoiner(dao._orm, dao.session().query(Agency), Stop.stop_name == 'FOOBAR').autojoin()
        self._check(query, ['routes', 'trips', 'stop_times', 'stops'])
        query = _AutoJoiner(dao._orm, dao.session().query(Agency), CalendarDate.date == '2016-01-01').autojoin()
        self._check(query, ['routes', 'trips', 'calendar', 'calendar_dates'])
        query = _AutoJoiner(dao._orm, dao.session().query(CalendarDate), Stop.stop_name == 'FOOBAR').autojoin()
        self._check(query, ['calendar', 'trips', 'stop_times', 'stops'])
        query = _AutoJoiner(dao._orm, dao.session().query(CalendarDate), Agency.agency_name == 'FOOBAR').autojoin()
        self._check(query, ['calendar', 'trips', 'routes', 'agency'])
        query = _AutoJoiner(dao._orm, dao.session().query(Stop), CalendarDate.date == '2016-01-01').autojoin()
        self._check(query, ['stop_times', 'trips', 'calendar', 'calendar_dates'])
        query = _AutoJoiner(dao._orm, dao.session().query(Stop), Agency.agency_name == 'FOOBAR').autojoin()
        self._check(query, ['stop_times', 'trips', 'routes', 'agency'])

        query = _AutoJoiner(dao._orm, dao.session().query(Agency), (Stop.stop_name == 'FOOBAR') & (CalendarDate.date == '2016-01-01')).autojoin()
        self._check(query, ['routes', 'trips', 'calendar', 'calendar_dates', 'stop_times', 'stops'])

        query = _AutoJoiner(dao._orm, dao.session().query(Trip), (Route.route_long_name == 'FOOBAR') & (StopTime.departure_time > 2000) & (CalendarDate.date == '2016-01-01')).autojoin()
        self._check(query, ['calendar', 'calendar_dates', 'routes', 'stop_times'])

        query = _AutoJoiner(dao._orm, dao.session().query(Trip), Agency.agency_name == 'FOOBAR').autojoin()
        self._check(query, ['routes', 'agency'])
        query = _AutoJoiner(dao._orm, dao.session().query(Trip), Stop.stop_name == 'FOOBAR').autojoin()
        self._check(query, ['stop_times', 'stops'])
        query = _AutoJoiner(dao._orm, dao.session().query(Trip), CalendarDate.date == '2016-01-01').autojoin()
        self._check(query, ['calendar', 'calendar_dates'])
        query = _AutoJoiner(dao._orm, dao.session().query(Route), Agency.agency_name == 'FOOBAR').autojoin()
        self._check(query, ['agency'])
        query = _AutoJoiner(dao._orm, dao.session().query(Route), Stop.stop_name == 'FOOBAR').autojoin()
        self._check(query, ['trips', 'stop_times', 'stops'])

        query = _AutoJoiner(dao._orm, dao.session().query(Shape), Agency.agency_name == 'FOOBAR').autojoin()
        self._check(query, ['trips', 'routes', 'agency'])

        query = _AutoJoiner(dao._orm, dao.session().query(FareAttribute), FareRule.contains_id == 'Z1').autojoin()
        self._check(query, ['fare_rules'])

    def _check(self, query, joins):
        query_joins = re.findall("JOIN\s+([a-z_]+)", str(query))
        # print(query_joins)
        self.assertTrue(joins == query_joins)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import math
import sys
import unittest
import six
from gtfsplugins.prettycsv import PrettyCsv

class TestPrettyPrinter(unittest.TestCase):

    def test_prettyprinter(self):
        # Capture standard output
        saved_stdout = sys.stdout
        try:
            out = six.StringIO()
            sys.stdout = out
            with PrettyCsv(None, fieldnames=[ 'col1', 'col2' ], maxwidth=5) as csv:
                csv.writerow({ 'col1': 1, 'col2': 2 })
                csv.writerow({ 'col2': 'foobarbaz', 'col1': 11 })
                csv.writerow([ 42, 'baz', 'extrawide' ])
            output1 = out.getvalue().strip()

            out = six.StringIO()
            sys.stdout = out
            with PrettyCsv(None, maxwidth=5) as csv:
                csv.writerow([ 1, 2 ])
                csv.writerow([ 11, 'foobarbaz', 'extrawide' ])
            output2 = out.getvalue().strip()

            out = six.StringIO()
            sys.stdout = out
            with PrettyCsv(None, fieldnames=[ 'col1', 'col2' ], maxwidth=5) as csv:
                csv.writerow([ 1 ])
                csv.writerow([ None, 1.42 ])
                csv.writerow([ None, 1./3., math.pi ])
            output3 = out.getvalue().strip()

        finally:
            sys.stdout = saved_stdout

        self.assertEqual("+------+-------+-------+\n"+
                         "| col1 |  col2 |       |\n"+
                         "+------+-------+-------+\n"+
                         "|    1 |     2 |       |\n"+
                         "|   11 | fooba |       |\n"+
                         "|   42 |   baz | extra |\n"+
                         "+------+-------+-------+", output1)

        self.assertEqual("+----+-------+-------+\n"+
                         "|  1 |     2 |       |\n"+
                         "| 11 | fooba | extra |\n"+
                         "+----+-------+-------+", output2)

        self.assertEqual("+------+-------+-------+\n"+
                         "| col1 |  col2 |       |\n"+
                         "+------+-------+-------+\n"+
                         "|    1 |       |       |\n"+
                         "|      |  1.42 |       |\n"+
                         "|      | 0.333 | 3.141 |\n"+
                         "+------+-------+-------+", output3)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from sqlalchemy.orm import clear_mappers

from gtfslib.spatial import orthodromic_distance
from gtfslib.dao import Dao

# Location of GTFS files to tests.
# This unit-test is not dependent on the content of a GTFS file.
# It just check for some generic assumptions that should be valid for any GTFS.
GTFS_LIST = [ "test/mini.gtfs.zip", "test/dummy.gtfs.zip", "test/sample-feed.zip" ]

DAO_URL = ""
# Set this to true to activate SQL logging
SQL_LOG = False

class TestGtfs(unittest.TestCase):

    def _test_one_gtfs(self, gtfs):
        clear_mappers()
        dao = Dao(DAO_URL, sql_logging=SQL_LOG)
        dao.load_gtfs(gtfs)

        # Check stop time normalization and interpolation
        for trip in dao.trips(prefetch_stop_times=True):
            stopseq = 0
            n_stoptimes = len(trip.stop_times)
            last_stop = None
            distance = trip.stop_times[0].shape_dist_traveled
            last_stoptime = None
            last_interpolated_speed = None
            for stoptime in trip.stop_times:
                self.assertTrue(stoptime.stop_sequence == stopseq)
                if stopseq == 0:
                    self.assertTrue(stoptime.arrival_time is None)
                else:
                    self.assertTrue(stoptime.arrival_time is not None)
                if stopseq == n_stoptimes - 1:
                    self.assertTrue(stoptime.departure_time is None)
                else:
                    self.assertTrue(stoptime.departure_time is not None)
                if last_stop is not None:
                    distance += orthodromic_distance(last_stop, stoptime.stop)
                last_stop = stoptime.stop
                if trip.shape is not None:
                    self.assertTrue(stoptime.shape_dist_traveled >= distance)
                else:
                    self.assertAlmostEqual(stoptime.shape_dist_traveled, distance, 1)
                stopseq += 1
                if stoptime.interpolated or (last_stoptime is not None and last_stoptime.interpolated):
                    dist = stoptime.shape_dist_traveled - last_stoptime.shape_dist_traveled
                    time = stoptime.arrival_time - last_stoptime.departure_time
                    speed = dist * 1.0 / time
                    if last_interpolated_speed is not None:
                        self.assertAlmostEqual(speed, last_interpolated_speed, 2)
                    last_interpolated_speed = speed
                if not stoptime.interpolated:
                    last_interpolated_speed = None
                last_stoptime = stoptime

        # Get all hops
        hops = dao.hops()
        nhops = 0
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 1 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)
            nhops += 1

        # Get hops with a delta of 2
        hops = dao.hops(delta=2)
        nhops2 = 0
        for st1, st2 in hops:
            self.assertTrue(st1.stop_sequence + 2 == st2.stop_sequence)
            self.assertTrue(st1.trip == st2.trip)
            nhops2 += 1
        ntrips = len(list(dao.trips()))
        # Assume all trips have len > 2
        self.assertTrue(nhops == nhops2 + ntrips)

        # Test shape_dist_traveled on stoptimes
        for trip in dao.trips():
            # Assume no shapes for now
            distance = 0.0
            last_stop = None
            for stoptime in trip.stop_times:
                if last_stop is not None:
                    distance += orthodromic_distance(last_stop, stoptime.stop)
                last_stop = stoptime.stop
                if trip.shape:
                    self.assertTrue(stoptime.shape_dist_traveled >= distance)
                else:
                    self.assertAlmostEqual(stoptime.shape_dist_traveled, distance, 2)

        # Test shape normalization
        for shape in dao.shapes():
            distance = 0.0
            last_pt = None
            ptseq = 0
            for point in shape.points:
                if last_pt is not None:
                    distance += orthodromic_distance(last_pt, point)
                last_pt = point
                self.assertAlmostEqual(point.shape_dist_traveled, distance, 2)
                self.assertTrue(point.shape_pt_sequence == ptseq)
                ptseq += 1

        # Check zone-stop relationship
        for zone in dao.zones(prefetch_stops=True):
            for stop in zone.stops:
                self.assertTrue(stop.zone == zone)
        for stop in dao.stops():
            if stop.zone:
                self.assertTrue(stop in stop.zone.stops)

    def test_all_gtfs(self):
        for gtfs in GTFS_LIST:
            print("Testing %s" % gtfs)
            self._test_one_gtfs(gtfs)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unicodedata
import shapefile
from gtfslib.spatial import SpatialClusterizer
from collections import defaultdict

class ShapefileExport(object):
    """
    Export data (stops, hops) to ESRI shapefile.
    Include frequency of use for each element
    * number of trips
    * number of trips x days
    Note: Usually the trips x days is a better usage metric as
    trips can be active on various number of days. If you want
    to compute the average number of trips per day, divide the
    result by the number of days (you can use date filtering).

    Parameters:
    --cluster=<dist>    Cluster stops closer than <dist> meters
    --stopshp=<file>    Output stops to given shapefile
    --hopshp=<file>     Output hops to given shapefile
    """

    def __init__(self):
        pass

    def remove_accents(self, strd):
        nfkd_form = unicodedata.normalize('NFKD', strd)
        only_ascii = nfkd_form.encode('ASCII', 'ignore')
        return only_ascii

    def run(self, context, stopshp=None, hopshp=None, cluster=0, **kwargs):
        cluster_meters = float(cluster)
        if stopshp is None and hopshp is None:
            print("Nothing to generate! Bailing out")
            return

        print("Loading stops...")
        stops = set()
        sc = SpatialClusterizer(cluster_meters)
        for stop in context.dao().stops(fltr=context.args.filter):
            sc.add_point(stop)
            stops.add(stop)
        print("Loaded %d stops. Clusterize..." % (len(stops)))
        sc.clusterize()
        print("Aggregated in %d clusters" % (len(sc.clusters())))

        print("Loading calendar dates")
        dates = set(context.dao().calendar_dates_date(fltr=context.args.filter))
        print("Loaded %d dates" % (len(dates)))

        print("Computing stop and hop trip count...")
        hop_tripcount = defaultdict(lambda: [0, 0])
        clu_tripcount = defaultdict(lambda: [0, 0])
        ntrips = 0
        for trip in context.dao().trips(fltr=context.args.filter, prefetch_stop_times=True, prefetch_stops=True, prefetch_calendars=True):
            # Compute the number of days the trip is running
            # RESTRICTED ON THE FILTERED DATES
            ndays = len([ date for date in trip.calendar.dates if date.as_date() in dates ])
            for st1, st2 in trip.hops():
                cluster1 = sc.cluster_of(st1.stop)
                cluster2 = sc.cluster_of(st2.stop)
                if cluster1 == cluster2:
                    pass
                key = (cluster1, cluster2)
                hop_tripcount[key][0] += 1
                hop_tripcount[key][1] += ndays
                clu_tripcount[cluster1][0] += 1
                clu_tripcount[cluster1][1] += ndays
            ntrips += 1
            if ntrips % 1000 == 0:
                print("%d trips..." % ntrips)

        if stopshp:
            print("Generating stops cluster shapefile...")
            stopshpwrt = shapefile.Writer(shapefile.POINT)
            stopshpwrt.field("id", "N")
            stopshpwrt.field("ids", "C", 100)
            stopshpwrt.field("name", "C", 200)
            stopshpwrt.field("ndep", "N")
            stopshpwrt.field("ndepday", "N")
            for cluster, (dep_count, depday_count) in clu_tripcount.items():
                stopshpwrt.point(cluster.lon(), cluster.lat()) # X,Y ?
                ids = cluster.aggregate(lambda s: s.stop_id, sep=';')
                names = cluster.aggregate(lambda s: s.stop_name, sep=';')
                stopshpwrt.record(cluster.id, self.remove_accents(ids),
                                  self.remove_accents(names),
                                  dep_count, depday_count)
            stopshpwrt.save(stopshp)

        if hopshp:
            print("Generating hop shapefile...")
            hopshpwrt = shapefile.Writer(shapefile.POLYLINE)
            hopshpwrt.field("from_id", "N")
            hopshpwrt.field("from_name", "C", 200)
            hopshpwrt.field("to_id", "N")
            hopshpwrt.field("to_name", "C", 200)
            hopshpwrt.field("name", "C", 200)
            hopshpwrt.field("ntrip", "N")
            hopshpwrt.field("ntripday", "N")
            for (c1, c2), (trip_count, tripday_count) in hop_tripcount.items():
                c1name = c1.aggregate(lambda s: s.stop_name, sep=';')
                c2name = c2.aggregate(lambda s: s.stop_name, sep=';')
                hopshpwrt.line(parts=[[[c1.lon(), c1.lat()], [c2.lon(), c2.lat()]]])
                hopshpwrt.record(c1.id, self.remove_accents(c1name), c2.id,
                                 self.remove_accents(c2name),
                                 self.remove_accents(c1name + " -> " + c2name),
                                 trip_count, tripday_count)
            hopshpwrt.save(hopshp)



# -*- coding: utf-8 -*-
#    This file is part of Gtfslib-python.
#
#    Gtfslib-python is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Gtfslib-python is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gtfslib-python.  If not, see <http://www.gnu.org/licenses/>.
"""
@author: Laurent GRÉGOIRE <laurent.gregoire@mecatran.com>
"""

import unittest

from gtfslib.dao import Dao
from sqlalchemy.orm import clear_mappers

# Location of broken.gtfs.zip.
# This unit-test is highly dependent on the CONTENT of this GTFS.
BROKEN_GTFS = "test/broken.gtfs.zip"

class TestMiniGtfs(unittest.TestCase):

    def test_broken(self):
        exception = False
        try:
            clear_mappers()
            dao = Dao("")
            dao.load_gtfs(BROKEN_GTFS, lenient=False)
        except KeyError:
            exception = True
        self.assertTrue(exception)

        clear_mappers()
        dao = Dao("")
        dao.load_gtfs(BROKEN_GTFS, lenient=True)

        # The following are based on BROKEN GTFS content,
        # that is the entities count minus broken ones.
        self.assertTrue(len(dao.routes()) == 4)
        self.assertTrue(len(list(dao.stops())) == 12)
        self.assertTrue(len(dao.calendars()) == 2)
        self.assertTrue(len(list(dao.trips())) == 104)
        self.assertTrue(len(dao.stoptimes()) == 500)
        self.assertTrue(len(dao.fare_attributes()) == 2)
        self.assertTrue(len(dao.fare_rules()) == 4)
        # This stop has missing coordinates in the broken file
        stop00 = dao.stop('FUR_CREEK_RES3')
        self.assertAlmostEquals(stop00.stop_lat, 0.0, 5)
        self.assertAlmostEquals(stop00.stop_lon, 0.0, 5)

if __name__ == '__main__':
    unittest.main()
